# -*- coding: utf-8 -*-
"""
Module to download ERA5 from terminal in netcdf and grib format.
"""
import warnings
import os
import logging
from datetime import datetime, time, timedelta
import shutil
import cdsapi
import numpy as np
import pandas as pd

from c3s_sm.misc import read_summary_yml

from repurpose.process import parallel_process
from repurpose.misc import delete_empty_directories

from ecmwf_models.utils import (
    lookup,
    update_image_summary_file,
    default_variables,
    split_array,
    check_api_ready,
    get_first_last_image_date
)
from ecmwf_models.extract import (
    save_ncs_from_nc,
    save_gribs_from_grib,
    unzip_nc,
)


def split_chunk(timestamps,
                n_vars,
                n_hsteps,
                max_req_size=1000,
                reduce=False,
                daily_request=False):
    """
    Split the passed time stamps into chunks for a valid request. One chunk
    can at most hold data for one month or one day, but cannot be larger than
    the maximum request size.

    Parameters
    ----------
    timestamps: pd.DatetimeIndex
        List of daily timestamps to split into chunks
    n_vars: int
        Number of variables in each request.
    max_req_size: int, optional (default: 1000)
        Maximum size of a request that the CDS API can handle
    reduce: bool, optional (default: False)
        Return only the start and end of each subperiod instead of all
        time stamps.
    daily_request: bool, optional (default: False)
        Only submit daily requests, otherwise monthly requests are allowed
        (if the max_req_size is not reached).

    Returns
    -------
    chunks: list
        List of start and end dates that contain a chunk that the API can
        handle.
    """
    n = int(max_req_size / n_vars / n_hsteps)

    def yield_chunk():
        for _, chunk_year in timestamps.groupby(timestamps.year).items():
            for _, chunk_month in chunk_year.groupby(chunk_year.month).items():
                if daily_request:
                    for _, chunk_day in chunk_month.groupby(
                            chunk_month.day).items():
                        yield chunk_day
                else:
                    yield chunk_month

    # each chunk contains either time stamps for one month, or for less,
    # if the request of one month would be too large.
    all_chunks = []
    for chunk in yield_chunk():
        if len(chunk) > n:
            chunks = split_array(chunk, n)
        else:
            chunks = np.array([chunk])
        for chunk in chunks:
            if reduce:
                all_chunks.append(np.array([chunk[0], chunk[-1]]))
            else:
                all_chunks.append(chunk)

    return all_chunks


def download_era5(
    c,
    years,
    months,
    days,
    h_steps,
    variables,
    target,
    grb=False,
    bbox=None,
    product="era5",
    dry_run=False,
    cds_kwds={},
):
    """
    Download era5 reanalysis data for single levels of a defined time span

    Parameters
    ----------
    c : cdsapi.Client
        Client to pass the request to
    years : list
        Years for which data is downloaded ,e.g. [2017, 2018]
    months : list
        Months for which data is downloaded, e.g. [4, 8, 12]
    days : list
        Days for which data is downloaded (range(31)=All days)
        e.g. [10, 20, 31]
    h_steps: list
        List of full hours to download data at the selected dates e.g [0, 12]
    variables : list, optional (default: None)
        List of variables to pass to the client, if None are passed, the
        default variables will be downloaded.
    target : str
        File name, where the data is stored.
    grb : bool, optional (default: False)
        Download data in grib format instead of netcdf
    bbox: Tuple[int,int,int,int], optional (default: None)
        Bounding box of the area to download
        (min_lon, min_lat, max_lon, max_lat) - wgs84.
        None will download global images.
    product : str
        ERA5 data product to download, either era5 or era5-land
    dry_run: bool, optional (default: False)
        Do not download anything, this is just used for testing the
        functionality
    cds_kwds: dict, optional
        Additional arguments to be passed to the CDS API retrieve request.

    Returns
    ---------
    success : bool
        Return True after downloading finished
    """

    if dry_run:
        return

    request = {
        "data_format": "grib" if grb else "netcdf",
        "download_format": "zip",
        "variable": variables,
        "year": [str(y) for y in years],
        "month": [str(m).zfill(2) for m in months],
        "day": [str(d).zfill(2) for d in days],
        "time": [time(h, 0).strftime("%H:%M") for h in h_steps],
    }

    if bbox is not None:   # maxlat, minlon, minlat, maxlon
        request["area"] = [bbox[3], bbox[0], bbox[1], bbox[2]]

    request.update(cds_kwds)
    if product == "era5":
        request["product_type"] = ["reanalysis"]
        c.retrieve("reanalysis-era5-single-levels", request, target)
    elif product == "era5-land":
        c.retrieve("reanalysis-era5-land", request, target)
    else:
        raise ValueError(
            product, "Unknown product, choose either 'era5' or 'era5-land'")

    return True


class CDSStatusTracker:
    """
    Track the status of the CDS download by using the CDS callback functions
    """

    statuscode_ok = 0
    statuscode_error = -1
    statuscode_unavailable = 10
    statuscode_terms_not_accepted = 11

    def __init__(self, logger=logging.getLogger()):
        self.download_statuscode = self.statuscode_ok
        self.logger = logger

    def handle_error_function(self, *args, **kwargs):
        if len(args) == 1:
            error_str = str(args[0])
            if '\n' in error_str:
                message_prefix, message_body = error_str.split('\n', 1)
            else:
                message_prefix = error_str
                message_body = ''
        else:
            message_prefix = args[0]
            message_body = args[1]

        if self.download_statuscode != self.statuscode_unavailable:
            if (message_prefix.startswith("Reason:") and
                    message_body == "Request returned no data"):
                self.download_statuscode = self.statuscode_unavailable
            elif ("Not all the required licences have been accepted" in
                  message_body):
                self.download_statuscode = self.statuscode_terms_not_accepted
            else:
                self.download_statuscode = self.statuscode_error

        self.logger.error(*args, **kwargs)


def download_and_move(
    target_path,
    startdate,
    enddate,
    product="era5",
    variables=None,
    keep_original=False,
    h_steps=(0, 6, 12, 18),
    grb=False,
    bbox=None,
    dry_run=False,
    grid=None,
    remap_method="bil",
    cds_kwds=None,
    stepsize="month",
    n_max_request=1000,
    keep_prelim=True,
    cds_token=None,
) -> int:
    """
    Downloads the data from the ECMWF servers and moves them to the target
    path.
    This is done in 30 day increments between start and end date.

    The files are then extracted into separate grib files per parameter and
    stored in yearly folders under the target_path.

    Parameters
    ----------
    target_path : str
        Path where the files are stored to
    startdate: datetime
        first date to download
    enddate: datetime
        last date to download
    product : str, optional (default: era5)
        Either era5 or era5-land
    variables : list, optional (default: None)
        Name of variables to download, see the documentation for all variable
        names. If None is chosen, then the 'default' variables are downloaded.
    keep_original: bool (default: False)
        If True, keep the original downloaded data stack as received from CDS
        after slicing individual time stamps.
    h_steps: tuple, optional (default: (0, 6, 12, 18))
        List of full hours to download data for at the selected dates e.g
        [0, 12] would download at 0:00 and 12:00. Only full hours are possible.
    grb: bool, optional (default: False)
        Download data as grib files instead of netcdf.
        Note that downloading in grib format, does not allow on-the-fly
        resampling (`grid` argument)
    bbox: Tuple[int,int,int,int], optional (default: None)
        Bounding box of the area to download
        (min_lon, min_lat, max_lon, max_lat) - wgs84.
        None will download global images.
    dry_run: bool
        Do not download anything, this is just used for testing the functions
    grid : dict, optional (default: None)
        A grid on which to remap the data using CDO. This must be a dictionary
        using CDO's grid description format, e.g.::

        grid = {
            "gridtype": "lonlat",
            "xsize": 720,
            "ysize": 360,
            "xfirst": -179.75,
            "yfirst": 89.75,
            "xinc": 0.5,
            "yinc": -0.5,
        }

        Default is to use no regridding.
        To use this option, it is necessary that CDO is installed.
    remap_method : str, optional (dafault: 'bil')
        Method to be used for regridding. Available methods are:
        - "bil": bilinear (default)
        - "bic": bicubic
        - "nn": nearest neighbour
        - "dis": distance weighted
        - "con": 1st order conservative remapping
        - "con2": 2nd order conservative remapping
        - "laf": largest area fraction remapping
    cds_kwds: dict, optional (default: None)
        Additional keyword arguments to be passed to the CDS API request.
        This might be useful in the future, when new server-side options are
        added which are not yet directly supported by this package.
    n_max_request : int, optional (default: 1000)
        Maximum size that a request can have to be processed by CDS. At the
        moment of writing this is 1000 (N_timstamps * N_variables in a request)
        but as this is a server side settings, it can change.
    keep_prelim: bool, optional (default: True)
        Keep preliminary data from ERA5T under a different file name.
        These data are not yet final and might change if an issue is detected.
        If False is chosen, then the preliminary data will be discarded and
        not stored.
    cds_token: str, optional (default: None)
        To identify with the CDS. Required if no .cdsapirc file exists in
        the home directory (see documentation). You can find your token/key
         on your CDS user profile page. Alternatively, the CDSAPI_KEY
         environment variable can be set manually instead of passing the token
         here.

    Returns
    -------
    status_code: int
        Status code summary from all requests:
        0 : All Downloaded data ok
        -1 : Error in at least one request
        -10 : No data available for requested time period
    """
    h_steps = list(h_steps)
    product = product.lower()
    cds_kwds = cds_kwds or dict()

    if variables is None:
        variables = default_variables(product=product)
    else:
        # find the dl_names
        variables = lookup(name=product, variables=variables)
        variables = variables["dl_name"].values.tolist()

    # this logger name is also used by CDS API, don't change it
    logger = logging.getLogger('cdsapi')

    if dry_run:
        warnings.warn("Dry run does not create connection to CDS")
        c = None
        cds_status_tracker = None
    else:
        if cds_token is not None:
            os.environ["CDSAPI_KEY"] = cds_token
        check_api_ready()

        os.makedirs(target_path, exist_ok=True)
        cds_status_tracker = CDSStatusTracker(logger=logger)
        c = cdsapi.Client(
            error_callback=cds_status_tracker.handle_error_function)

    timestamps = pd.DatetimeIndex(
        np.array([
            datetime(t.year, t.month, t.day)
            for t in pd.date_range(startdate, enddate, freq='D')
        ]))

    req_periods = split_chunk(
        timestamps,
        n_vars=len(variables),
        n_hsteps=len(h_steps),
        max_req_size=n_max_request,
        reduce=True,
        daily_request=True if stepsize == "day" else False)

    logger.info(f"Request is split into {len(req_periods)} chunks")
    logger.info(f"Target directory {target_path}")

    downloaded_data_path = os.path.join(target_path, "temp_downloaded")
    if not os.path.exists(downloaded_data_path):
        os.mkdir(downloaded_data_path)

    def _download(curr_start, curr_end):
        curr_start = pd.to_datetime(curr_start).to_pydatetime()
        curr_end = pd.to_datetime(curr_end).to_pydatetime()

        status_code = -1

        fname = "{start}_{end}.{ext}".format(
            start=curr_start.strftime("%Y%m%d"),
            end=curr_end.strftime("%Y%m%d"),
            ext="zip" if grb is False else "grb")

        dl_file = os.path.join(downloaded_data_path, fname)

        finished, i = False, 0

        while (not finished) and (i < 5):  # try max 5 times
            try:
                finished = download_era5(
                    c,
                    years=[curr_start.year],
                    months=[curr_start.month],
                    days=range(curr_start.day, curr_end.day + 1),
                    h_steps=h_steps,
                    variables=variables,
                    grb=grb,
                    bbox=bbox,
                    product=product,
                    target=dl_file,
                    dry_run=dry_run,
                    cds_kwds=cds_kwds,
                )
                status_code = 0
                break

            except Exception as e:  # noqa: E722
                c.error(e)

                # If no data is available or terms were not accepted we don't
                # need to retry
                if (cds_status_tracker.download_statuscode in [
                        CDSStatusTracker.statuscode_unavailable,
                        CDSStatusTracker.statuscode_terms_not_accepted
                ]):
                    status_code = -10
                    break

                # delete the partly downloaded data and retry
                if os.path.isfile(dl_file):
                    os.remove(dl_file)
                finished = False
                i += 1
                continue

        if status_code == 0 and os.path.exists(dl_file):
            if grb:
                save_gribs_from_grib(
                    dl_file,
                    target_path,
                    product_name=product.upper(),
                    keep_original=keep_original,
                    keep_prelim=keep_prelim)
            else:
                # Extract and merge nc files from zip
                dl_file_new = dl_file.replace('.zip', '.nc')
                unzip_nc(dl_file, dl_file_new)
                dl_file = dl_file_new

                save_ncs_from_nc(
                    dl_file,
                    target_path,
                    product_name=product.upper(),
                    grid=grid,
                    remap_method=remap_method,
                    keep_original=keep_original,
                    keep_prelim=keep_prelim)

        return status_code

    # Since we download month/month or day/day we need to
    # collect all the status codes to return a valid
    # status code for the whole time period
    all_status_codes = parallel_process(
        _download,
        ITER_KWARGS={
            'curr_start': [p[0] for p in req_periods],
            'curr_end': [p[1] for p in req_periods]
        },
        logger_name='cdsapi',
        loglevel='DEBUG',
        n_proc=1,
        backend='multiprocessing')

    # remove temporary files
    if not keep_original:
        shutil.rmtree(downloaded_data_path)

    if grid is not None:
        gridpath = os.path.join(target_path, "grid.txt")
        if os.path.exists(gridpath):
            os.unlink(gridpath)
        weightspath = os.path.join(target_path, "remap_weights.nc")
        if os.path.exists(weightspath):
            os.unlink(weightspath)

    delete_empty_directories(target_path)

    dl_settings = {
        'product': product,
        'variables': variables,
        'keep_original': keep_original,
        'h_steps': h_steps,
        'grb': grb,
        'bbox': bbox,
        'grid': grid,
        'remap_method': remap_method,
        'cds_kwds': cds_kwds,
        'stepsize': stepsize,
        'n_max_request': n_max_request,
        'keep_prelim': keep_prelim,
    }

    update_image_summary_file(target_path, dl_settings)

    handlers = logger.handlers[:]

    for handler in handlers:
        logger.removeHandler(handler)
        handler.close()
    handlers.clear()

    # if any of the sub-periods was successful we want the function to return 0
    consolidated_status_code = max(all_status_codes)
    return consolidated_status_code


def download_record_extension(path, dry_run=False, cds_token=None):
    """
    Uses information from an existing record to download additional data
    from CDS.

    Parameters
    ----------
    path: str
        Path where the image data to extend is stored. Must also contain
        a `summary.yml` file.
    dry_run: bool, optional
        Do not download anything, this is just used for testing the functions
    cds_token: str, optional (default: None)
        To identify with the CDS. Required if no `.cdsapirc` file exists in
        the home directory (see documentation). You can find your token/key
        on your CDS user profile page. Alternatively, the CDSAPI_KEY
        environment variable can be set manually instead of passing the token
        here.

    Returns
    -------
    status_code: int
        Status code summary from all requests:
        0 : All Downloaded data ok
        -1 : Error in at least one request
        -10 : No data available for requested time period
    """
    props = read_summary_yml(path)

    last_img = get_first_last_image_date(path)

    startdate = pd.to_datetime(last_img).to_pydatetime() + timedelta(days=1)

    enddate = (pd.to_datetime(datetime.now().date()).to_pydatetime()
               - timedelta(days=1))  # yesterday

    logging.info(f"Downloading record extension from {startdate} to {enddate}")
    logging.info(f"Additional settings {props['download_settings']}")

    return download_and_move(
        path,
        startdate=startdate,
        enddate=enddate,
        cds_token=cds_token,
        dry_run=dry_run,
        **props['download_settings']
    )



# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2016, TU Wien
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Test module for image to time series conversion.
"""

import os
import glob
import tempfile
import numpy as np
import numpy.testing as nptest
import subprocess
import shutil
from pathlib import Path
import xarray as xr
import yaml
from datetime import datetime

from c3s_sm.misc import read_summary_yml

from ecmwf_models.era5.reshuffle import Reshuffler
from ecmwf_models import ERATs

inpath = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    "..",
    "ecmwf_models-test-data",
    "ERA5",
)

def test_cli_reshuffle_and_update():
    testdata_path = Path(os.path.join(inpath, 'netcdf'))
    with tempfile.TemporaryDirectory() as tempdir:
        tempdir = Path(tempdir)
        img_path = tempdir / 'images'
        shutil.copytree(testdata_path, img_path)

        # we duplicate 1 file to check whether prioritizing final images over T images works
        ds = xr.open_dataset(testdata_path / '2010' / '001' / "ERA5_AN_20100101_0000.nc")
        ds['swvl1'].values = np.full(ds['swvl1'].values.shape, 99)
        ds.to_netcdf(img_path / '2010' / '001' / "ERA5-T_AN_20100101_0000.nc")

        ts_path = tempdir / 'ts'

        subprocess.call(["era5", "reshuffle", str(img_path), str(ts_path),
                         "--end", "2010-01-01", "-v", "swvl1,swvl2", "-l",
                         "True", "--bbox", "12.0", "46.0", "17.0", "50.0",
                         "--h_steps", "0"])

        ts_reader = ERATs(str(ts_path))
        ts = ts_reader.read(15, 48)
        assert 99 not in ts['swvl1'].values  # verify ERA5-T was NOT used!
        swvl1_values_should = np.array([0.402825], dtype=np.float32)
        nptest.assert_allclose(
            ts["swvl1"].values, swvl1_values_should, rtol=1e-5)

        ts_reader.close()

        # Manipulate settings to update with different time stamp for same day
        props = read_summary_yml(str(ts_path))
        props['img2ts_kwargs']['h_steps'] = [12]
        props['img2ts_kwargs']['startdate'] = datetime(2009, 12, 31)
        props['img2ts_kwargs']['enddate'] = datetime(2009, 12, 31)

        with open(ts_path / 'overview.yml', 'w') as f:
            yaml.dump(props, f, default_flow_style=False, sort_keys=False)

        subprocess.call(["era5", "update_ts", str(ts_path)])
        ts_reader = ERATs(str(ts_path))
        ts = ts_reader.read(15, 48)
        swvl1_values_should = np.array([0.402825, 0.390983], dtype=np.float32)
        nptest.assert_allclose(
            ts["swvl1"].values, swvl1_values_should, rtol=1e-5)

        assert 'swvl2' in ts.columns

        ts_reader.close()

def test_ERA5_reshuffle_nc():
    # test reshuffling era5 netcdf images to time series

    startdate = "2010-01-01"
    enddate = "2010-01-01"
    parameters = ["swvl1", "swvl2"]
    h_steps = [0, 12]
    landpoints = True
    bbox = (12, 46, 17, 50)

    with tempfile.TemporaryDirectory() as ts_path:
        reshuffler = Reshuffler(os.path.join(inpath, 'netcdf'), ts_path,
                                variables=parameters, h_steps=h_steps,
                                land_points=landpoints)
        reshuffler.reshuffle(startdate, enddate, bbox=bbox)

        assert (len(glob.glob(os.path.join(ts_path, "*.nc"))) == 5)
        # less files because only land points and bbox
        ds = ERATs(ts_path, ioclass_kws={"read_bulk": True})
        ts = ds.read(15, 48)
        ds.close()
        swvl1_values_should = np.array([0.402825, 0.390983], dtype=np.float32)
        nptest.assert_allclose(
            ts["swvl1"].values, swvl1_values_should, rtol=1e-5)
        swvl2_values_should = np.array([0.390512, 0.390981], dtype=np.float32)
        nptest.assert_allclose(
            ts["swvl2"].values, swvl2_values_should, rtol=1e-5)


def test_ERA5_reshuffle_grb():
    # test reshuffling era5 netcdf images to time series

    startdate = "2010-01-01"
    enddate = "2010-01-01"
    parameters = ["swvl1", "swvl2"]
    h_steps = [0, 12]
    landpoints = False
    bbox = (12, 46, 17, 50)

    with tempfile.TemporaryDirectory() as ts_path:

        reshuffler = Reshuffler(os.path.join(inpath, 'grib'), ts_path,
                                variables=parameters, h_steps=h_steps,
                                land_points=landpoints)
        reshuffler.reshuffle(startdate, enddate, bbox=bbox)

        assert len(glob.glob(os.path.join(ts_path, "*.nc"))) == 5
        ds = ERATs(ts_path, ioclass_kws={"read_bulk": True})
        ts = ds.read(15, 48)
        ds.close()
        swvl1_values_should = np.array([0.402824, 0.390979], dtype=np.float32)
        nptest.assert_allclose(
            ts["swvl1"].values, swvl1_values_should, rtol=1e-5)
        swvl2_values_should = np.array([0.390514, 0.390980], dtype=np.float32)
        nptest.assert_allclose(
            ts["swvl2"].values, swvl2_values_should, rtol=1e-5)



# -*- coding: utf-8 -*-
'''
Testing the utility functions
'''

import os
import tempfile
import numpy as np
from netCDF4 import Dataset

from ecmwf_models.utils import (
    parse_filetype,
    parse_product,
    load_var_table,
    lookup,
    get_default_params,
    make_era5_land_definition_file,
)


def test_parse_product():
    inpath = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "ecmwf_models-test-data",
        "ERA5", "netcdf")
    assert parse_product(inpath) == 'era5'


def test_parse_filetype():
    inpath = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "ecmwf_models-test-data",
        "ERA5", "netcdf")
    assert parse_filetype(inpath) == 'netcdf'


def test_load_var_table():
    table = load_var_table('era5')
    assert table.index.size == 260
    assert table.loc[
        100].dl_name == 'mean_surface_downward_short_wave_radiation_flux'
    assert table.loc[100].short_name == 'msdwswrf'

    table = load_var_table('era5-land')
    assert table.index.size == 50
    assert table.loc[46].dl_name == 'volumetric_soil_water_layer_1'
    assert table.loc[46].short_name == 'swvl1'


def test_lookup():
    lut = lookup('era5', ['swvl1', 'stl1'])
    assert lut.loc[249].dl_name == 'volumetric_soil_water_layer_1'
    assert lut.loc[166].dl_name == 'soil_temperature_level_1'


def test_get_default_params():
    vars = get_default_params('era5')
    assert all(vars['default'].values == 1)


def test_create_land_definition_file():
    inpath = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), "ecmwf_models-test-data",
        "ERA5", "netcdf")
    path_desired = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), '..', 'src',
        'ecmwf_models', 'era5', 'land_definition_files',
        'landmask_0.25_0.25.nc')

    data_file = os.path.join(inpath, '2010', '001', 'ERA5_AN_20100101_0000.nc')
    out_dir = tempfile.mkdtemp()
    out_file = os.path.join(out_dir, 'land_definition_out.nc')

    make_era5_land_definition_file(data_file, out_file)

    ds_actual = Dataset(out_file)
    actual = ds_actual.variables['land'][:]

    ds_desired = Dataset(path_desired)
    desired = ds_desired.variables['land'][:]

    assert np.allclose(actual, desired, equal_nan=True)



"""
Tests actually downloading data from the CDS
"""

import os
import tempfile
import unittest
from datetime import datetime
import subprocess

from ecmwf_models.era5.img import ERA5NcImg
from ecmwf_models.utils import check_api_ready

try:
    check_api_ready()
    api_ready = True
except ValueError:
    api_ready = False

class DownloadTest(unittest.TestCase):

    # these tests only run if a username and pw are set in the environment
    # variables. To manually set them: `export USERNAME="my_username"` etc.

    @unittest.skipIf(
        os.environ.get('CDSAPI_KEY') is None and not api_ready,
        'CDSAPI_KEY not found.'
    )
    def test_cli_download(self):

        with tempfile.TemporaryDirectory() as dl_path:
            startdate = enddate = "2023-01-01"

            args = [
                dl_path, '-s', startdate, '-e', enddate,
                '-v', 'swvl1', '--h_steps', '0',
                '--bbox', "-10", "30", "60", "90",
            ]

            subprocess.call(['era5', 'download'] + args)

            # # For debugging don't use subprocess
            # from ecmwf_models.era5.download import download_and_move
            # download_and_move(dl_path, startdate, enddate,
            #                   variables=['swvl1'],
            #                   h_steps=[0], bbox=[-10, 30, 60, 90])

            out_path = os.path.join(dl_path, '2023', '001')
            assert (os.path.exists(out_path))
            imgs = os.listdir(out_path)
            assert len(imgs) == 1

            ds = ERA5NcImg(os.path.join(out_path, imgs[0]), parameter='swvl1')
            img = ds.read(datetime(2023, 1, 1))
            assert img.data['swvl1'].shape == (241, 281)



# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2019, TU Wien
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Utility functions for all data products in this package.
"""
import os
import warnings
from datetime import datetime
import pandas as pd
import numpy as np
from netCDF4 import Dataset
from collections import OrderedDict
from parse import parse
import yaml

from ecmwf_models.extract import save_gribs_from_grib
from repurpose.misc import find_first_at_depth

from ecmwf_models.globals import (DOTRC, CDS_API_URL, IMG_FNAME_TEMPLATE,
                                  IMG_FNAME_DATETIME_FORMAT,
                                  SUPPORTED_PRODUCTS)


def parse_product(inpath: str) -> str:
    """
    Tries to find out what product is stored in the path. This is done based
    on the name of the first file in the path that is found.

    Parameters
    ----------
    inpath: str
        Input path where ERA data was downloaded to. Contains annual folders.

    Returns
    -------
    product : str
        Product name
    """
    props = img_infer_file_props(inpath)

    if "era5-land" in props['product'].lower():
        return "era5-land"  # also era5-land-t
    elif "era5" in props['product'].lower():
        return "era5"  # also era5-t
    else:
        raise ValueError(
            f"Could not derive product name from data in {inpath}")


def parse_filetype(inpath: str) -> str:
    """
    Tries to find out the file type by parsing filenames in the passed
    directory.

    Parameters
    ----------
    inpath: str
        Input path where ERA data was downloaded to. Contains annual folders.

    Returns
    -------
    product : str
        Product name
    """
    props = img_infer_file_props(inpath)
    if props['ext'] == 'grb':
        return 'grib'
    else:
        return 'netcdf'


def load_var_table(name="era5", lut=False):
    """
    Load the variables table for supported variables to download.

    Parameters
    ----------
    lut : bool, optional (default: False)
        If set to true only names are loaded, so that they can be used
        for a LUT otherwise the full table is loaded
    """
    name = name.lower()
    if name == "era5":
        era_vars_csv = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), "era5", "era5_lut.csv")
    elif name == "era5-land":
        era_vars_csv = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "era5",
            "era5-land_lut.csv",
        )
    else:
        raise ValueError(name, "No LUT for the selected dataset found.")

    if lut:
        dat = pd.read_csv(era_vars_csv)[["dl_name", "long_name", "short_name"]]
    else:
        dat = pd.read_csv(era_vars_csv)

    return dat


def lookup(name, variables):
    """
    Search the passed elements in the lookup table, if one does not exists,
    print a Warning
    """
    lut = load_var_table(name=name, lut=True)

    selected = []
    for var in variables:
        found = False
        for row in lut.itertuples():
            if var in row:
                selected.append(row.Index)
                found = True
                break
        if found:
            continue
        else:
            raise ValueError(
                f"Passed variable {var} is not a supported variable.")

    return lut.loc[selected, :]


def get_default_params(name="era5"):
    """
    Read only lines that are marked as default variable in the csv file

    Parameters
    ----------
    name : str
        Name of the product to get the default parameters for
    """
    vars = load_var_table(name, lut=False)
    return vars.loc[vars.default == 1.0]


def default_variables(product="era5", format='dl_name'):
    """
    These variables are being downloaded, when None are passed by the user

    Parameters
    ---------
    product : str, optional (default: 'era5')
        Name of the era5 product to read the default variables for.
        Either 'era5' or 'era5-land'.
    format: str, optional (default: 'dl_name')
        'dl_name' for name as in the downloaded image data
        'short_name' for short name
        'long_name' for long name
    """
    lut = load_var_table(name=product)
    defaults = lut.loc[lut["default"] == 1][format].values
    return defaults.tolist()


def make_era5_land_definition_file(
    data_file,
    out_file,
    data_file_y_res=0.25,
    ref_var="lsm",
    threshold=0.5,
    exclude_antarctica=True,
):
    """
    Create a land grid definition file from a variable within a downloaded,
    regular (netcdf) era5 file.

    Parameters
    ----------
    data_file : str
        Path to the downloaded file that contains the image that is used as the
        reference for creating the land definition file.
    out_file: str
        Full output path to the land definition file to create.
    data_file_y_res : float, optional (default: 0.25)
        The resolution of the data file in latitude direction.
    ref_var: str, optional (default: 'lsm')
        A variable in the data_file that is the reference for the
        land definition.
        By default, we use the land-sea-mask variable.
    threshold: float, optional (default: 0.5)
        Threshold value below which a point is declared water,
        and above (or equal) which it is declared a land-point.
        If None is passed, then a point is declared a land point
        if it is not masked (numpy masked array) in the reference variable.
    exclude_antarctica: bool, optional (default: True)
        Cut off the definition file at -60° Lat to exclude Land Points
        in Antarctica.
    """
    lat_name, lon_name = "latitude", "longitude"
    ds_in = Dataset(data_file)
    ds_out = Dataset(out_file, "w", format="NETCDF4")

    for dim_name in ds_in.dimensions.keys():
        ds_out.createDimension(dim_name, size=ds_in.dimensions[dim_name].size)
        ds_out.createVariable(dim_name, "float32", (dim_name,), zlib=True)
        ds_out.variables[dim_name][:] = ds_in.variables[dim_name][:]

    ref = ds_in.variables[ref_var]

    land_mask = np.zeros(ref.shape)

    if np.isnan(threshold):
        land_mask[~ref[:].mask] = 1.0
    else:
        land_mask[ref[:] >= threshold] = 1.0

    # drop values below -60° Lat
    if exclude_antarctica:
        cut_off_lat = -60.0
        index_thres_lat = ((180.0 / data_file_y_res) + 1) - (
            (90.0 + cut_off_lat) / data_file_y_res)
        land_mask[int(index_thres_lat):, :] = np.nan
    else:
        cut_off_lat = None

    ds_out.createVariable("land", "float32", (lat_name, lon_name), zlib=True)
    ds_out.variables["land"][:] = land_mask

    land_attrs = OrderedDict([
        ("units", "(0,1)"),
        ("long_name", "Land-sea mask"),
        ("based_on_variable", ref_var),
        ("standard_name", "land_binary_mask"),
        ("threshold_land_>=", str(threshold)),
        ("cut_off_at", str(cut_off_lat)),
    ])

    for attr, val in land_attrs.items():
        ds_out.variables["land"].setncattr(attr, val)

    ds_in.close()
    ds_out.close()


def split_array(array, chunk_size):
    """
    Split an array into chunks of a given size.

    Parameters
    ----------
    array : array-like
        Array to split into chunks
    chunk_size : int
        Size of each chunk

    Returns
    -------
    chunks : list
        List of chunks
    """
    chunks = []
    for i in range(0, len(array), chunk_size):
        chunks.append(array[i:i + chunk_size])
    return chunks


def check_api_ready() -> bool:
    """
    Verify that the API is ready to be used. Otherwise raise an Error.

    Returns:
    --------
    api_ready: bool
        True if api is ready
    """
    if not os.path.isfile(DOTRC):
        key = os.environ.get('CDSAPI_KEY')
        if "CDSAPI_URL" not in os.environ:
            os.environ['CDSAPI_URL'] = CDS_API_URL

        if key is None:
            raise ValueError(
                'Neither CDSAPI_KEY variable nor .cdsapirc file found, '
                'download will not work! '
                'Please create a .cdsapirc file with your API key. '
                'See: https://cds.climate.copernicus.eu/how-to-api')
        else:
            return True
    else:
        if "CDSAPI_URL" in os.environ:
            os.environ.pop("CDSAPI_URL")  # Use URL from file
        return True


def img_infer_file_props(img_root_path: str,
                         fntempl: str = IMG_FNAME_TEMPLATE,
                         start_from_last=False) -> dict:
    """
    Parse file names to retrieve properties from fntempl.
    Does not open any files.

    Parameters
    ----------
    img_root_path: str
        Root directory where annual directories are located
    fntempl: str, optional
        Filename template to parse filenames with
    start_from_last: bool, optional
        Use the last available file instead of the first one.
    """
    fname = find_first_at_depth(img_root_path, 2, reverse=start_from_last)

    if fname is None:
        raise ValueError(f"No matching files for chosen template found in "
                         f"the directory {img_root_path}")
    else:
        file_args = parse(fntempl, fname)
        return file_args.named


def get_first_last_image_date(path, start_from_last=True):
    """
    Parse files in the given directory (or any subdir) using the passed
    filename template. props will contain all fields specified in the template.
    the `datetime` field is required and used to determine the last image date.

    Parameters
    ----------
    path: str
        Path to the directory containing the image files
    start_from_last: bool, optional (default: True')
        Get date from last available file instead of the first available one.

    Returns
    -------
    date: str
        Parse date from the last found image file that matches `fntempl`.
    """
    try:
        props = img_infer_file_props(
            path, fntempl=IMG_FNAME_TEMPLATE, start_from_last=start_from_last)
        dt = datetime.strptime(props['datetime'], IMG_FNAME_DATETIME_FORMAT)
        dt = datetime(dt.year, dt.month, dt.day)
    except ValueError:
        raise ValueError('Could not infer date from filenames. '
                         'Check filename template.')

    return str(dt)


def update_image_summary_file(data_path: str,
                              other_props: dict = None,
                              out_file=None):
    """
    Summarize image metadata as yml file

    Parameters
    ----------
    data_path: str
        Root path to the image archive
    other_props: dict, optional (default: None)
        Other properties to write into the yml file. E.g. download
        options to enable time series update.
    out_file: str, optional (default: None)
        Path to summary file. File will be created/updated.
        If not specified, then `data_path` is used. If a file already exists,
        it will be overwritten.
    """
    try:
        first_image_date = get_first_last_image_date(
            data_path, start_from_last=False)
        last_image_date = get_first_last_image_date(
            data_path, start_from_last=True)
    except ValueError:
        warnings.warn(f"Could not infer date from filenames in {data_path}")
        return

    props = img_infer_file_props(data_path, start_from_last=False)
    _ = props.pop("datetime")
    props['period_from'] = str(pd.to_datetime(first_image_date).date())
    props['period_to'] = str(pd.to_datetime(last_image_date).date())

    props['last_update'] = str(datetime.now())

    props['download_settings'] = {}

    if other_props is not None:
        for k, v in other_props.items():
            props['download_settings'][k] = v

    if out_file is None:
        out_file = os.path.join(data_path, f"overview.yml")

    with open(out_file, 'w') as f:
        yaml.dump(props, f, default_flow_style=False)


def assert_product(product: str) -> str:
    if product not in SUPPORTED_PRODUCTS:
        raise ValueError(f"Got product {product} but expected one of "
                         f"{SUPPORTED_PRODUCTS}")
    return product



# -*- coding: utf-8 -*-
"""
This module contains ERA5/ERA5-Land specific child classes of the netcdf
and grib base classes, that are used for reading all ecmwf products.
"""

from typing import Optional, Collection
from typing_extensions import Literal
from pygeogrids.grids import CellGrid

from ecmwf_models.utils import assert_product
from ecmwf_models.interface import ERANcImg, ERANcDs, ERAGrbImg, ERAGrbDs


class ERA5NcImg(ERANcImg):

    def __init__(
        self,
        filename: str,
        parameter: Collection[str] = None,
        product: Literal['era5', 'era5-land'] = 'era5',
        subgrid: Optional[CellGrid] = None,
        mask_seapoints: Optional[bool] = False,
        array_1D: Optional[bool] = False,
    ):
        """
        Reader for a single ERA5 netcdf image file.

        Parameters
        ----------
        filename: str
            Path to the image file to read.
        parameter: list[str] or str, optional (default: None)
            Name of parameters to read from the image file.
            None means all parameter
        product: str, optional (default: 'era5')
            What era5 product, either era5 or era5-land.
        subgrid: pygeogrids.CellGrid, optional (default: None)
            Read only data for points of this grid and not global values.
        mask_seapoints : bool, optional (default: False)
            Read the land-sea mask to mask points over water and
            set them to nan. This option needs the 'lsm' parameter to be
            in the file!
        array_1D: bool, optional (default: False)
            Read data as list, instead of 2D array, used for reshuffling.
        """

        super(ERA5NcImg, self).__init__(
            filename=filename,
            product=assert_product(product.lower()),
            parameter=parameter,
            subgrid=subgrid,
            mask_seapoints=mask_seapoints,
            array_1D=array_1D,
        )


class ERA5NcDs(ERANcDs):
    """
    Reader for a stack of ERA5 netcdf image files.

    Parameters
    ----------
    root_path: str
        Path to the image files to read.
    parameter: list[str] or str, optional (default: None)
        Name of parameters to read from the image file.
        None means all parameter
    product: str, optional (default: 'era5')
        What era5 product, either `era5` or `era5-land`.
    h_steps : list, optional (default: (0,6,12,18))
        List of full hours to read images for.
    subgrid: pygeogrids.CellGrid, optional (default: None)
        Read only data for points of this grid and not global values.
    mask_seapoints : bool, optional (default: False)
        Read the land-sea mask to mask points over water and set them to nan.
        This option needs the 'lsm' parameter to be in the file!
    array_1D: bool, optional (default: False)
        Read data as list, instead of 2D array, used for reshuffling.
    """

    def __init__(
        self,
        root_path: str,
        parameter: Collection[str] = None,
        product: str = 'era5',
        h_steps: Collection[int] = (0, 6, 12, 18),
        subgrid: Optional[CellGrid] = None,
        mask_seapoints: Optional[bool] = False,
        array_1D: Optional[bool] = False,
    ):
        super(ERA5NcDs, self).__init__(
            root_path=root_path,
            product=assert_product(product.lower()),
            parameter=parameter,
            subgrid=subgrid,
            h_steps=h_steps,
            array_1D=array_1D,
            mask_seapoints=mask_seapoints,
        )


class ERA5GrbImg(ERAGrbImg):

    def __init__(
        self,
        filename: str,
        parameter: Collection[str] = None,
        subgrid: Optional[CellGrid] = None,
        mask_seapoints: Optional[bool] = False,
        array_1D=False,
    ):
        """
        Reader for a single ERA5 grib image file.

        Parameters
        ----------
        filename: str
            Path to the image file to read.
        parameter: list[str] or str, optional (default: None)
            Name of parameters to read from the image file.
            None means all parameter
        subgrid: pygeogrids.CellGrid, optional (default: None)
            Read only data for points of this grid and not global values.
        mask_seapoints : bool, optional (default: False)
            Read the land-sea mask to mask points over water and set
            them to nan. This option needs the 'lsm' parameter to be in
            the file!
        array_1D: bool, optional (default: False)
            Read data as list, instead of 2D array, used for reshuffling.
        """
        super(ERA5GrbImg, self).__init__(
            filename=filename,
            product="era5",
            parameter=parameter,
            subgrid=subgrid,
            mask_seapoints=mask_seapoints,
            array_1D=array_1D,
        )


class ERA5GrbDs(ERAGrbDs):

    def __init__(
        self,
        root_path: str,
        parameter: Collection[str] = None,
        h_steps: Collection[int] = (0, 6, 12, 18),
        product: str = "era5",
        subgrid: Optional[CellGrid] = None,
        mask_seapoints: Optional[bool] = False,
        array_1D: Optional[bool] = False,
    ):
        """
        Reader for a stack of ERA5 grib image file.

        Parameters
        ----------
        root_path: str
            Path to the image files to read.
        parameter: list[str] or str, optional (default: None)
            Name of parameters to read from the image file.
            None means all parameter
        h_steps : list, optional (default: [0,6,12,18])
            List of full hours to read images for.
        product: str, optional (default: 'era5')
            What era5 product, either `era5` or `era5-land`.
        subgrid: pygeogrids.CellGrid, optional (default: None)
            Read only data for points of this grid and not global values.
        mask_seapoints : bool, optional (default: False)
            Read the land-sea mask to mask points over water and set them
            to nan. This option needs the 'lsm' parameter to be in the file!
        array_1D: bool, optional (default: False)
            Read data as list, instead of 2D array, used for reshuffling.
        """

        super(ERA5GrbDs, self).__init__(
            root_path=root_path,
            product=assert_product(product.lower()),
            parameter=parameter,
            subgrid=subgrid,
            h_steps=h_steps,
            mask_seapoints=mask_seapoints,
            array_1D=array_1D,
        )



# The MIT License (MIT)
#
# Copyright (c) 2019, TU Wien, Department of Geodesy and Geoinformation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Base classes for reading downloaded ERA netcdf and grib images and stacks
"""
import warnings
import os
import glob
from datetime import timedelta, datetime  # noqa: F401
import numpy as np
import xarray as xr

from pygeobase.io_base import ImageBase, MultiTemporalImageBase
from pygeobase.object_base import Image
from pynetcf.time_series import GriddedNcOrthoMultiTs
from pygeogrids.grids import BasicGrid, gridfromdims
from pygeogrids.netcdf import load_grid

from ecmwf_models.grid import trafo_lon
from ecmwf_models.utils import lookup
from ecmwf_models.globals import (
    IMG_FNAME_TEMPLATE,
    IMG_FNAME_DATETIME_FORMAT,
    SUPPORTED_PRODUCTS,
    SUBDIRS,
)
from ecmwf_models.globals import (pygrib, pygrib_available,
                                  PygribNotFoundError)


class ERANcImg(ImageBase):
    """
    Reader for a single ERA netcdf file. The main purpose of this class is
    to use it in the time series conversion routine. To read downloaded image
    files, we recommend using xarray (https://docs.xarray.dev/en/stable/).

    Parameters
    ----------
    filename: str
        Path to the image file to read.
    product : str
        'era5' or 'era5-land'
    parameter: list or str, optional (default: ['swvl1', 'swvl2'])
        Name of parameters to read from the image file.
    subgrid: ERA_RegularImgGrid or ERA_RegularImgLandGrid or None, optional
        Read only data for points of this grid.
        If None is passed, we read all points from the file.
        The main purpose of this parameter is when reshuffling to time series,
        to include only e.g. points over land.
    mask_seapoints : bool, optional (default: False)
        Read the land-sea mask to mask points over water and set them to nan.
        This option needs the 'lsm' parameter to be in the file!
    array_1D: bool, optional (default: False)
        Read data as list, instead of 2D array, used for reshuffling.
    mode : str, optional (default: 'r')
        Mode in which to open the file, changing this can cause data loss.
        This argument should not be changed!
    """

    def __init__(
        self,
        filename,
        product,
        parameter=None,
        subgrid=None,
        mask_seapoints=False,
        array_1D=False,
        mode='r',
    ):

        super(ERANcImg, self).__init__(filename, mode=mode)

        if parameter is not None:
            # look up short names
            self.parameter = lookup(
                product, np.atleast_1d(parameter))["short_name"].values
        else:
            self.parameter = None

        self.mask_seapoints = mask_seapoints
        self.array_1D = array_1D
        self.subgrid = subgrid

        if self.subgrid and not self.array_1D:
            warnings.warn(
                "Reading spatial subsets as 2D arrays ony works if there "
                "is an equal number of points in each line")

    def read(self, timestamp=None):
        """
        Read data from the loaded image file.

        Parameters
        ---------
        timestamp : datetime, optional (default: None)
            Specific date (time) to read the data for.
        """

        try:
            dataset = xr.open_dataset(
                self.filename, engine="netcdf4", mask_and_scale=True)
        except IOError as e:
            print(" ".join([self.filename, "can not be opened"]))
            raise e

        if self.parameter is None:
            self.parameter = list(dataset.data_vars)

        if self.mask_seapoints:
            if "lsm" not in dataset.variables.keys():
                raise IOError("No land sea mask parameter (lsm) in"
                              " passed image for masking.")
            else:
                sea_mask = dataset.variables["lsm"].values
        else:
            sea_mask = None

        return_img = {}
        return_metadata = {}

        grid = gridfromdims(
            trafo_lon(dataset['longitude'].values),
            dataset['latitude'].values,
            origin='top')

        if self.subgrid is not None:
            gpis = grid.find_nearest_gpi(self.subgrid.activearrlon,
                                         self.subgrid.activearrlat)[0]
        else:
            gpis = None

        for name in self.parameter:
            try:
                variable = dataset[name]
            except KeyError:
                path, f = os.path.split(self.filename)
                warnings.warn(f"Cannot load variable {name} from file {f}. "
                              f"Filling image with NaNs.")
                dat = np.full(
                    grid.shape if gpis is None else len(gpis), np.nan)
                return_img[name] = dat
                continue

            if 'expver' in variable.dims and (variable.data.ndim == 3):
                warnings.warn(f"Found experimental data in {self.filename}")
                param_data = variable.data[-1]
                for vers_data in variable.data:
                    if not np.all(np.isnan(vers_data)):
                        param_data = vers_data
            else:
                param_data = variable.data

            if self.mask_seapoints:
                param_data = np.ma.array(
                    param_data,
                    mask=np.logical_not(sea_mask),
                    fill_value=np.nan,
                )
                param_data = param_data.filled()

            param_data = param_data.flatten()

            if gpis is not None:
                param_data = param_data[gpis]

            return_metadata[name] = variable.attrs
            return_img[name] = param_data

        dataset.close()

        if self.subgrid is None:
            self.subgrid = grid

        if self.array_1D:
            return Image(
                self.subgrid.activearrlon,
                self.subgrid.activearrlat,
                return_img,
                return_metadata,
                timestamp,
            )
        else:
            if len(self.subgrid.shape) != 2:
                raise ValueError("Passed subgrid does not have a 2d shape."
                                 "Did you mean to read data as 1d arrays?")

            for key in return_img:
                return_img[key] = return_img[key].reshape(self.subgrid.shape)

            return Image(
                self.subgrid.activearrlon.reshape(self.subgrid.shape),
                self.subgrid.activearrlat.reshape(self.subgrid.shape),
                return_img,
                return_metadata,
                timestamp,
            )

    def write(self, data):
        raise NotImplementedError()

    def flush(self):
        pass

    def close(self):
        pass


class ERANcDs(MultiTemporalImageBase):
    """
    Reader to extract individual images from a multi-image netcdf dataset.
    The main purpose of this class is to use it in the time series conversion
    routine. To read downloaded image files, we recommend using
    xarray (https://docs.xarray.dev/en/stable/).

    Parameters
    ----------
    root_path: str
        Root path where image data is stored.
    product: str
        ERA5 or ERA5-LAND
    parameter: list[str] or str, optional (default: None)
        Parameter or list of parameters to read. None reads all available
        Parameters.
    subgrid: ERA_RegularImgGrid or ERA_RegularImgLandGrid, optional
        Read only data for points of this grid.
        If None is passed, we read all points from the file.
        The main purpose of this parameter is when reshuffling to time series,
        to include only e.g. points over land.
    mask_seapoints: bool, optional (default: False)
        All points that are not over land are replaced with NaN values.
        This requires that the land sea mask (lsm) parameter is included
        in the image files!
    h_steps: tuple, optional (default: (0, 6, 12, 18))
        Time stamps available for each day. Numbers refer to full hours.
    array_1D: bool, optional (default: True)
        Read data as 1d arrays. This is required when the passed subgrid
        is 1-dimensional (e.g. when only landpoints are read). Otherwise
        when a 2d (subgrid) is used, this switch means that the extracted
        image data is also 2-dimensional (lon, lat).
    """

    def __init__(
            self,
            root_path,
            product,
            parameter=None,
            subgrid=None,
            mask_seapoints=False,
            h_steps=(0, 6, 12, 18),
            array_1D=False,
    ):

        self.h_steps = h_steps

        if parameter is not None:
            # look up short names
            self.parameter = lookup(
                product, np.atleast_1d(parameter))["short_name"].values
        else:
            self.parameter = None

        ioclass_kws = {
            'product': product,
            'parameter': parameter,
            'subgrid': subgrid,
            'mask_seapoints': mask_seapoints,
            'array_1D': array_1D
        }

        # the goal is to use ERA5-T*.nc if necessary, but prefer ERA5*.nc
        self.fn_templ_priority = [
            IMG_FNAME_TEMPLATE.format(
                product=(p + ext).upper(),
                type='*',
                datetime="{datetime}",
                ext='nc') for ext in ['', '-T'] for p in SUPPORTED_PRODUCTS
        ]

        super(ERANcDs, self).__init__(
            root_path,
            ERANcImg,
            fname_templ=IMG_FNAME_TEMPLATE.format(
                product='*', type='*', datetime='{datetime}', ext='nc'),
            datetime_format=IMG_FNAME_DATETIME_FORMAT,
            subpath_templ=SUBDIRS,
            exact_templ=False,
            ioclass_kws=ioclass_kws)

    def _search_files(self,
                      timestamp,
                      custom_templ=None,
                      str_param=None,
                      custom_datetime_format=None):
        """
        override the original filename generation to allow multiple files for
        time stamp
        """
        if custom_templ is not None:
            raise NotImplementedError
        else:
            fname_templ = self.fname_templ

        if custom_datetime_format is not None:
            dFormat = {self.dtime_placeholder: custom_datetime_format}

        else:
            dFormat = {self.dtime_placeholder: self.datetime_format}

        sub_path = ''
        if self.subpath_templ is not None:
            for s in self.subpath_templ:
                sub_path = os.path.join(sub_path, timestamp.strftime(s))

        fname_templ = fname_templ.format(**dFormat)

        if str_param is not None:
            fname_templ = fname_templ.format(**str_param)

        search_file = os.path.join(self.path, sub_path,
                                   timestamp.strftime(fname_templ))

        if self.exact_templ:
            raise NotImplementedError
        else:
            filename = glob.glob(search_file)
            if len(filename) > 1:
                for templ in self.fn_templ_priority:
                    fname_templ = templ.format(**dFormat)
                    if str_param is not None:
                        fname_templ = fname_templ.format(**str_param)
                    search_file = os.path.join(self.path, sub_path,
                                               timestamp.strftime(fname_templ))
                    filename = glob.glob(search_file)
                    if len(filename) == 1:
                        break

        if not filename:
            filename = []

        return filename

    def tstamps_for_daterange(self, start_date, end_date):
        """
        Get datetimes in the correct sub-daily resolution between 2 dates

        Parameters
        ----------
        start_date: datetime
            Start datetime
        end_date: datetime
            End datetime

        Returns
        ----------
        timestamps : list
            List of datetimes
        """

        img_offsets = np.array([timedelta(hours=h) for h in self.h_steps])

        timestamps = []
        diff = end_date - start_date
        for i in range(diff.days + 1):
            daily_dates = start_date + timedelta(days=i) + img_offsets
            timestamps.extend(daily_dates.tolist())

        return timestamps


class ERAGrbImg(ImageBase):

    def __init__(self,
                 filename,
                 product,
                 parameter=None,
                 subgrid=None,
                 mask_seapoints=False,
                 array_1D=True,
                 mode='r'):
        """
        Reader for a single ERA grib file. The main purpose of this class is
        to use it in the time series conversion routine. To read downloaded image
        files, we recommend using xarray (https://docs.xarray.dev/en/stable/).

        Parameters
        ----------
        filename: str
            Path to the image file to read.
        product : str
            ERA5 or ERA5-LAND
        parameter: list or str, optional
            Name of parameters to read from the image file. None means all
            available data variables.
        subgrid: ERA_RegularImgGrid or ERA_RegularImgLandGrid, optional (default: None)
            Read only data for points of this grid.
            If None is passed, we read all points from the file.
        mask_seapoints : bool, optional (default: False)
            Read the land-sea mask to mask points over water and set them to nan.
            This option needs the 'lsm' parameter to be in the file!
        array_1D: bool, optional (default: False)
            Read data as list, instead of 2D array, used for reshuffling.
        mode : str, optional (default: 'r')
            Mode in which to open the file, changing this can cause data loss.
            This argument should not be changed!
        """
        super(ERAGrbImg, self).__init__(filename, mode=mode)

        if parameter is None:
            self.parameter = None
        else:
            # look up short names
            self.parameter = lookup(
                product, np.atleast_1d(parameter))["short_name"].values

        self.product = product

        self.mask_seapoints = mask_seapoints
        self.array_1D = array_1D
        self.subgrid = subgrid

    def read(self, timestamp=None):
        """
        Read data from the loaded image file.

        Parameters
        ---------
        timestamp : datetime, optional (default: None)
            Specific date (time) to read the data for.
        """
        if not pygrib_available:
            raise PygribNotFoundError()
        grbs = pygrib.open(self.filename)

        return_img = {}
        return_metadata = {}

        grid = None

        for n in range(1, grbs.messages + 1):
            message = grbs.message(n)
            try:
                param_name = str(message.cfVarNameECMF)  # old field?
            except RuntimeError:
                param_name = str(message.shortName)

            if self.mask_seapoints and (param_name == "lsm"):
                pass
            elif self.parameter is None:
                pass
            elif param_name in self.parameter:
                pass
            else:
                continue

            return_metadata[param_name] = {}

            message = grbs.message(n)
            lats, lons = message.latlons()
            param_data = message.values

            if grid is None:
                grid = BasicGrid(
                    trafo_lon(lons).flatten(),
                    lats.flatten(),
                    shape=param_data.shape)

            param_data = param_data.flatten()

            if self.subgrid is not None:
                gpis = grid.find_nearest_gpi(self.subgrid.activearrlon,
                                             self.subgrid.activearrlat)[0]

                param_data = param_data[gpis]

            return_img[param_name] = param_data

            return_metadata[param_name]["units"] = message["units"]
            return_metadata[param_name]["long_name"] = \
                message["parameterName"]

            if "levels" in message.keys():
                return_metadata[param_name]["depth"] = "{:} cm".format(
                    message["levels"])

        grbs.close()

        # Set data for non-land points to NaN
        if self.mask_seapoints:
            if 'lsm' not in return_img:
                raise IOError(
                    "No land sea mask parameter (lsm) in passed image"
                    " for masking.")
            else:
                # mask the loaded data
                mask = np.logical_not(return_img['lsm'].flatten())
                for name in return_img.keys():
                    param_data = return_img[name]
                    param_data = np.ma.array(
                        param_data,
                        mask=mask,
                        fill_value=np.nan,
                    )
                    param_data = param_data.filled()
                    return_img[name] = param_data

            if (self.parameter is not None) and ('lsm' not in self.parameter):
                return_img.pop('lsm')
                return_metadata.pop('lsm')

        if self.subgrid is None:
            self.subgrid = grid

        # Add empty arrays for missing variables
        if self.parameter is not None:
            for p in self.parameter:
                if p not in return_img:
                    param_data = np.full(np.prod(self.subgrid.shape), np.nan)
                    warnings.warn(
                        f"Cannot load variable {param_name} from file "
                        f"{self.filename}. Filling image with NaNs.")
                    return_img[param_name] = param_data
                    return_metadata[param_name] = {}
                    return_metadata[param_name]["long_name"] = lookup(
                        self.product, [param_name]).iloc[0]["long_name"]

        if self.array_1D:
            return Image(
                self.subgrid.activearrlon,
                self.subgrid.activearrlat,
                return_img,
                return_metadata,
                timestamp,
            )
        else:
            if len(self.subgrid.shape) != 2:
                raise ValueError("Passed subgrid does not have a 2d shape."
                                 "Did you mean to read data as 1d arrays?")

            for key in return_img:
                return_img[key] = return_img[key].reshape(self.subgrid.shape)

            return Image(
                self.subgrid.activearrlon.reshape(self.subgrid.shape),
                self.subgrid.activearrlat.reshape(self.subgrid.shape),
                return_img,
                return_metadata,
                timestamp,
            )

    def write(self, data):
        raise NotImplementedError()

    def flush(self):
        pass

    def close(self):
        pass


class ERAGrbDs(MultiTemporalImageBase):

    def __init__(
            self,
            root_path,
            product,
            parameter=None,
            subgrid=None,
            mask_seapoints=False,
            h_steps=(0, 6, 12, 18),
            array_1D=True,
    ):
        """
        Reader to extract individual images from a multi-image grib dataset.
        The main purpose of this class is to use it in the time series conversion
        routine. To read downloaded image files, we recommend using
        xarray (https://docs.xarray.dev/en/stable/).

        Parameters
        ----------
        root_path: str
            Root path where the downloaded image data is stored in grib format.
            We assume that image files are organized in subfolders by year,
            with each year containing subfolders for individual days of the
            year.
        product : str
            ERA5 or ERA5-Land
        parameter: list[str] or str, optional (default: None)
            Parameter or list of parameters to read. None reads all available
            Parameters.
        subgrid: ERA_RegularImgGrid or ERA_RegularImgLandGrid, optional
            Read only data for points of this grid.
            If None is passed, we read all points from the file.
        mask_seapoints: bool, optional (default: False)
            All points that are not over land are replaced with NaN values.
            This requires that the land sea mask (lsm) parameter is included
            in the image files!
        h_steps: tuple, optional (default: (0, 6, 12, 18))
            Time stamps available for each day. Numbers refer to full hours.
        array_1D: bool, optional (default: True)
            Read data as 1d arrays. This is required when the passed subgrid
            is 1-dimensional (e.g. when only landpoints are read). Otherwise
            when a 2d (subgrid) is used, this switch means that the extracted
            image data is also 2-dimensional (lon, lat).
        """
        self.h_steps = h_steps

        ioclass_kws = {
            "product": product,
            "parameter": parameter,
            "subgrid": subgrid,
            "mask_seapoints": mask_seapoints,
            "array_1D": array_1D,
        }

        fname_templ = IMG_FNAME_TEMPLATE.format(
            product="*", type="*", datetime="{datetime}", ext="grb")

        super(ERAGrbDs, self).__init__(
            root_path,
            ERAGrbImg,
            fname_templ=fname_templ,
            datetime_format=IMG_FNAME_DATETIME_FORMAT,
            subpath_templ=SUBDIRS,
            exact_templ=False,
            ioclass_kws=ioclass_kws,
        )

    def tstamps_for_daterange(self, start_date, end_date):
        """
        Get datetimes in the correct sub-daily resolution between 2 dates

        Parameters
        ----------
        start_date: datetime
            Start datetime
        end_date: datetime
            End datetime

        Returns
        ----------
        timestamps : list
            List of datetime values (between start and end date) for all
            required time stamps.
        """
        img_offsets = np.array([timedelta(hours=h) for h in self.h_steps])

        timestamps = []
        diff = end_date - start_date
        for i in range(diff.days + 1):
            daily_dates = start_date + timedelta(days=i) + img_offsets
            timestamps.extend(daily_dates.tolist())

        return timestamps


class ERATs(GriddedNcOrthoMultiTs):
    """
    Time series reader for all reshuffled ERA reanalysis products in time
    series format (pynetcf OrthoMultiTs format)
    Use the read_ts(lon, lat) resp. read_ts(gpi) function of this class
    to read data for a location!
    """

    def __init__(self, ts_path, grid_path=None, **kwargs):
        """
        Parameters
        ----------
        ts_path : str
            Directory where the netcdf time series files are stored
        grid_path : str, optional (default: None)
            Path to grid file, that is used to organize the location of time
            series to read. If None is passed, grid.nc is searched for in the
            ts_path.

        Optional keyword arguments that are passed to the Gridded Base when used
        ------------------------------------------------------------------------
        parameters : list, optional (default: None)
            Specific variable names to read, if None are selected,
            all are read.
        offsets : dict, optional (default: None)
            Offsets (values) that are added to the parameters (keys)
        scale_factors : dict, optional (default: None)
            Offset (value) that the parameters (key) is multiplied with
        ioclass_kws: dict, (optional)

        Optional keyword arguments, passed to the OrthoMultiTs class when used
        ----------------------------------------------------------------------
            read_bulk : boolean, optional (default: False)
                If set to True, the data of all locations is read into memory,
                and subsequent calls to read_ts then read from cache and
                not from disk. This makes reading complete files faster.
            read_dates : boolean, optional (default: False)
                If false, dates will not be read automatically but only on
                specific request useable for bulk reading because currently
                the netCDF num2date routine is very slow for big datasets.
        """
        if grid_path is None:
            grid_path = os.path.join(ts_path, "grid.nc")
        grid = load_grid(grid_path)

        super(ERATs, self).__init__(ts_path, grid, **kwargs)



import click
from datetime import datetime

from ecmwf_models.era5.download import (download_and_move,
                                        download_record_extension)
from ecmwf_models.utils import (default_variables)

from ecmwf_models.era5.reshuffle import Reshuffler, extend_ts


@click.command(
    "download",
    context_settings={
        'show_default': True,
        'help_option_names': ['-h', '--help']
    },
    short_help="Download ERA5 reanalysis image data between two "
    "dates from the Copernicus Climate Data Store (CDS). "
    "Before this program can be used, you have to "
    "register an account at CDS, and should setup a `.cdsapirc` "
    "file as described here: "
    "https://cds.climate.copernicus.eu/how-to-api")
@click.argument("PATH", type=click.Path(writable=True))
@click.option(
    '--start',
    '-s',
    type=click.STRING,
    default=str(datetime(1979, 1, 1)),
    help="First date of the period to download data for. Format: YYYY-MM-DD")
@click.option(
    '--end',
    '-e',
    type=click.STRING,
    default=str(datetime.now().date()),
    help="Last date of the period to download data for. Format: YYYY-MM-DD")
@click.option(
    '--variables',
    '-v',
    type=click.STRING,
    default=','.join(default_variables('era5', 'short_name')),
    help="Name of variables to download. To download multiple variables pass "
    "comma-separated names (e.g. `-v swvl1,stl1`). "
    "For all available variables see the docs. ")
@click.option(
    "-k",
    "--keep_original",
    type=click.BOOL,
    default=False,
    help="To keep the original image stacks as downloaded from CDS, "
    "instead of deleting them after extracting individual images, pass "
    "`--keep_original True`")
@click.option(
    "-grb",
    "--as_grib",
    type=click.BOOL,
    default=False,
    help="To download data in grib format instead of netcdf pass "
    "`--as_grib True`.")
@click.option(
    "--h_steps",
    type=str,
    default="0,6,12,18",
    help="Temporal sampling of downloaded images. To download multiple time "
    "stamps or each day, pass comma-separated values. "
    "Pass a set of full hours here, like '--h_steps 0,12' to download "
    "two images for each day, at 0:00 and 12:00 respectively. "
    "By default, we download 6-hourly images starting at 0:00 UTC, "
    "(i.e. `--h_steps 0,6,12,18`")
@click.option(
    '--bbox',
    nargs=4,
    type=click.FLOAT,
    default=None,
    help="4 NUMBERS | min_lon min_lat max_lon max_lat. "
    "Set Bounding Box (lower left and upper right corner) "
    "of area to download (WGS84). Default is global.")
@click.option(
    "--keep_prelim",
    type=click.BOOL,
    default=True,
    help="The last 1-2 month of data are usually 'preliminary' (labelled as "
    "'ERA5-T' and 'ERA5-Land-T') and might be changed if an issue is "
    "detected. When this option is deactivated (`--keep_prelim False`), "
    "only the final data will be kept, while the ERA5-T data is discarded "
    "after download. By default, we also keep preliminary files, but they "
    "get a different file name as the final data.")
@click.option(
    "--max_request_size",
    type=int,
    default=1000,
    help="Maximum number of requests to pass to the CDS API. "
    "The default is 1000, but what is allowed, depends on server "
    "settings. Server settings may change at some point. Change "
    "accordingly here in case that 'the request is too large'. "
    "A smaller number will results in smaller download chunks (slower).")
@click.option(
    "--cds_token",
    type=click.STRING,
    default=None,
    help="To identify with the CDS. Required only if no `.cdsapirc` file "
    "exists in the home directory (see documentation). "
    "You can find your token/key on your CDS user profile page. "
    "Alternatively, you can also set an environment variable "
    "`CDSAPI_KEY` with your token.")
def cli_download_era5(path, start, end, variables, keep_original, as_grib,
                      h_steps, bbox, keep_prelim, max_request_size, cds_token):
    """
    Download ERA5 image data within the chosen period. NOTE: Before using this
    program, create a CDS account and set up a `.cdsapirc` file as described
    here: https://cds.climate.copernicus.eu/how-to-api

    \b
    Required Parameters
    -------------------
    > PATH: string (required)
          Root of local filesystem where the downloaded data will be stored.
          Make sure to set up the CDS API for your account as described at
          https://cds.climate.copernicus.eu/how-to-api
    """
    # The docstring above is slightly different to the normal python one to
    # display it properly on the command line.

    h_steps = [int(h.strip()) for h in h_steps.split(',')]
    variables = [str(v.strip()) for v in variables.split(',')]

    status_code = download_and_move(
        target_path=path,
        startdate=start,
        enddate=end,
        product="era5",
        variables=variables,
        h_steps=h_steps,
        grb=as_grib,
        bbox=bbox,
        keep_original=keep_original,
        stepsize='month',
        n_max_request=max_request_size,
        keep_prelim=keep_prelim,
        cds_token=cds_token,
    )

    return status_code


@click.command(
    "download",
    context_settings={
        'show_default': True,
        'help_option_names': ['-h', '--help']
    },
    short_help="Download ERA5 reanalysis image data between two "
    "dates from the Copernicus Climate Data Store (CDS). "
    "Before this program can be used, you have to "
    "register an account at CDS, and should setup a `.cdsapirc` "
    "file as described here: "
    "https://cds.climate.copernicus.eu/how-to-api")
@click.argument("PATH", type=click.Path(writable=True))
@click.option(
    '--start',
    '-s',
    type=click.STRING,
    default=str(datetime(1979, 1, 1)),
    help="First date of the period to download data for. Format: YYYY-MM-DD")
@click.option(
    '--end',
    '-e',
    type=click.STRING,
    default=str(datetime.now().date()),
    help="Last date of the period to download data for. Format: YYYY-MM-DD")
@click.option(
    '--variables',
    '-v',
    type=click.STRING,
    default=','.join(default_variables('era5-land',
                                       'short_name')),
    help="Name of variables to download. To download multiple variables pass "
    "comma-separated names (e.g. `-v swvl1,stl1`). "
    "For all available variables see the docs. ")
@click.option(
    "-k",
    "--keep_original",
    type=click.BOOL,
    default=False,
    help="To keep the original image stacks as downloaded from CDS, "
    "instead of deleting them after extracting individual images, pass "
    "`--keep_original True`")
@click.option(
    "-grb",
    "--as_grib",
    type=click.BOOL,
    default=False,
    help="To download data in grib format instead of netcdf pass "
    "`--as_grib True`.")
@click.option(
    "--h_steps",
    type=click.STRING,
    default='0,6,12,18',
    help="Temporal sampling of downloaded images. To download multiple time "
    "stamps or each day, pass comma-separated values. "
    "Pass a set of full hours here, like '--h_steps 0,12' to download "
    "two images for each day, at 0:00 and 12:00 respectively. "
    "By default, we download 6-hourly images starting at 0:00 UTC, "
    "(i.e. `--h_steps 0,6,12,18`")
@click.option(
    "--keep_prelim",
    type=click.BOOL,
    default=True,
    help="The last 1-2 month of data are usually 'preliminary' (labelled as "
    "'ERA5-T' and 'ERA5-Land-T') and might be changed if an issue is "
    "detected. When this option is deactivated (`--keep_prelim False`), "
    "only the final data will be kept, while the ERA5-T data is discarded "
    "after download. By default, we also keep preliminary files, but they "
    "get a different file name as the final data.")
@click.option(
    "--max_request_size",
    type=int,
    default=1000,
    help="Maximum number of requests to pass to the CDS API. "
    "The default is 1000, but what is allowed, depends on server "
    "settings. Server settings may change at some point. Change "
    "accordingly here in case that 'the request is too large'. "
    "A smaller number will results in smaller download chunks (slower).")
@click.option(
    "--cds_token",
    type=click.STRING,
    default=None,
    help="To identify with the CDS. Required only if no `.cdsapirc` file "
    "exists in the home directory (see documentation). "
    "You can find your token/key on your CDS user profile page. "
    "Alternatively, you can also set an environment variable "
    "`CDSAPI_KEY` with your token.")
def cli_download_era5land(path, start, end, variables, keep_original, as_grib,
                          h_steps, keep_prelim, max_request_size, cds_token):
    """
    Download ERA5-Land image data within a chosen period.
    NOTE: Before using this program, create a CDS account and set up a
    `.cdsapirc` file as described here:
    https://cds.climate.copernicus.eu/how-to-api

    \b
    Required Parameters
    -------------------
    > PATH: string (required)
          Root of local filesystem where the downloaded data will be stored.
          Make sure to set up the CDS API for your account as described at
          https://cds.climate.copernicus.eu/how-to-api
    """
    # The docstring above is slightly different to the normal python one to
    # display it properly on the command line.

    h_steps = [int(h.strip()) for h in h_steps.split(',')]
    variables = [str(v.strip()) for v in variables.split(',')]

    status_code = download_and_move(
        target_path=path,
        startdate=start,
        enddate=end,
        product="era5-land",
        variables=variables,
        h_steps=h_steps,
        grb=as_grib,
        keep_original=keep_original,
        stepsize='month',
        n_max_request=max_request_size,
        keep_prelim=keep_prelim,
        cds_token=cds_token)

    return status_code


@click.command(
    "update_img",
    context_settings={
        'show_default': True,
        'help_option_names': ['-h', '--help']
    },
    short_help="Extend an existing set of images by downloading new data "
    "with the same settings as before."
    "NOTE: First use the `era5 download` or `era5land download` "
    "programs.")
@click.argument("path", type=click.Path(writable=True))
@click.option(
    "--cds_token",
    type=click.STRING,
    default=None,
    help="To identify with the CDS. Required only if no `.cdsapirc` file "
    "exists in the home directory (see documentation). "
    "You can find your token/key on your CDS user profile page. "
    "Alternatively, you can also set an environment variable "
    "`CDSAPI_KEY` with your token.")
def cli_update_img(path, cds_token=None):
    """
    Download new images from CDS to your existing local archive. Use the same
    settings as before.
    NOTE: First use the `era5 download` or `era5land download` programs.
    The so-created archive can then be updated using this program.

    \b
    Required Parameters
    -------------------
    > PATH: string (required)
          Path where previously downloaded images are stored. There must
          be a file `overview.yml` which contains the download settings.
          Make sure to set up the CDS API for your account as described at
          https://cds.climate.copernicus.eu/how-to-api,
    """
    # The docstring above is slightly different to the normal python one to
    # display it properly on the command line.

    download_record_extension(path, cds_token=cds_token)


@click.command(
    "reshuffle",
    context_settings={
        'show_default': True,
        'help_option_names': ['-h', '--help']
    },
    short_help="Convert previously downloaded ERA5/ERA5-Land image "
    "data into a time series format.")
@click.argument("IMG_PATH", type=click.Path(readable=True))
@click.argument("TS_PATH", type=click.Path(writable=True))
@click.option(
    '--start',
    '-s',
    type=click.STRING,
    default=None,
    help="First date of the period to reshuffle data for. By default, the "
         "first available image is used. Format: YYYY-MM-DD")
@click.option(
    '--end',
    '-e',
    type=click.STRING,
    default=None,
    help="Last date of the period to reshuffle data for. By default, the last"
         "available image is used. Format: YYYY-MM-DD")
@click.option(
    '--variables',
    '-v',
    type=click.STRING,
    default=None,
    help="Subset of variables to convert. Pass comma-separated names"
    "to select multiple variables (short names, as in the input images, "
    "e.g. `-v swvl1,stl1`). If not specified, all variables from the "
    "first image file are used.")
@click.option(
    '--land_points',
    '-l',
    type=click.BOOL,
    show_default=True,
    default=False,
    help="To store only time series for points that are over land, pass "
         "`--land_points True`. ")
@click.option(
    '--bbox',
    nargs=4,
    type=click.FLOAT,
    default=None,
    help="4 NUMBERS | min_lon min_lat max_lon max_lat. "
    "Set Bounding Box (lower left and upper right corner) "
    "of area to reshuffle (WGS84). Default is global.")
@click.option(
    "--h_steps",
    type=click.STRING,
    default="0,6,12,18",
    help="Full hour time stamps of images to include in time series. "
    "To select multiple, pass comma-separated values here, "
    "e.g. '--h_steps 0,12' will only include data from images at "
    "0:00 and 12:00 UTC and ignore all other available time stamps."
    "By default, we include data for every 6th hour each day.")
@click.option(
    "--cellsize",
    type=click.FLOAT,
    default=5.0,
    help="Chunking for the output time series. By default, we create 5x5 "
         "degree time series files."
)
@click.option(
    '--imgbuffer',
    '-b',
    type=click.INT,
    default=50,
    help="Number of images to read into memory at once before "
    "conversion to time series. A larger buffer means faster "
    "processing but requires more memory.")
def cli_reshuffle(img_path, ts_path, start, end, variables, land_points, bbox,
                  h_steps, cellsize, imgbuffer):
    """
    Convert previously downloaded ERA5 or ERA5-Land image data into a
    time series format.

    \b
    Required Parameters
    -------------------
    > IMG_PATH: string (required)
          Directory where the downloaded image data is stored, i.e., where
          annual folders are found.
    > TS_PATH: string (required)
          Root of local filesystem where the time series should be stored.
    """

    h_steps = [int(h.strip()) for h in h_steps.split(',')]

    if variables is not None:
        variables = [str(v.strip()) for v in variables.split(',')]

    print(f"Converting ERA5/ERA5-Land images in period from "
          f"{'first available image' if start is None else start} to "
          f"{'last available image' if end is None else end} into "
          f"time series to {ts_path}.")

    reshuffler = Reshuffler(img_path, ts_path,
                            variables=variables,
                            h_steps=h_steps,
                            land_points=land_points,
                            product=None  # Infer prod automatically from files
                            )
    reshuffler.reshuffle(startdate=start, enddate=end, bbox=bbox,
                         cellsize=cellsize, imgbuffer=imgbuffer)


@click.command(
    "update_ts",
    context_settings={
        'show_default': True,
        'help_option_names': ['-h', '--help']
    },
    short_help="Append new image data to an existing time series archive.")
@click.argument("TS_PATH", type=click.Path(writable=True))
@click.option(
    '--imgpath',
    '-p',
    type=click.STRING,
    default=None,
    help="Manually specify where the image data to extend the time "
    "series are located. If this is not specified, we use the previously "
    "used path from the `overview.yml` file stored with the time series "
    "data.")
@click.option(
    '--imgbuffer',
    '-b',
    type=click.INT,
    default=50,
    help="Number of images to read into memory at once before "
    "conversion to time series. A larger buffer means faster "
    "processing but requires more memory.")
def cli_extend_ts(ts_path, imgpath, imgbuffer):
    """
    Append new image data to an existing time series archive. The archive must
    be created first using the `reshuffle` program. We will use the same
    settings as in the initial run (see `overview.yml` in TS_PATH) for
    consistent extensions.

    \b
    Required Parameters
    -------------------
    > TS_PATH: string (required)
          Root of local filesystem where the time series from a previous call
          of `reshuffle` are stored. New data will be appended to the existing
          files!
    """
    kwargs = dict(ts_path=ts_path, imgbuffer=imgbuffer)

    if imgpath is not None:  # otherwise use path from yml
        kwargs["input_root"] = imgpath

    extend_ts(**kwargs)


@click.group(short_help="ERA5 Command Line Programs imported from the "
             "`ecmwf_models` pip package.")
def era5():
    pass


era5.add_command(cli_download_era5)
era5.add_command(cli_update_img)
era5.add_command(cli_reshuffle)
era5.add_command(cli_extend_ts)


@click.group(short_help="ERA5-Land Command Line Programs imported from the "
             "`ecmwf_models` pip package.")
def era5land():
    pass


era5land.add_command(cli_download_era5land)
era5land.add_command(cli_update_img)
era5land.add_command(cli_reshuffle)
era5land.add_command(cli_extend_ts)



import tempfile
from datetime import datetime
import logging
import os
import pandas as pd
import xarray as xr
import zipfile
import shutil
import numpy as np

from ecmwf_models.globals import (IMG_FNAME_TEMPLATE,
                                  IMG_FNAME_DATETIME_FORMAT, EXPVER, SUBDIRS)
from ecmwf_models.globals import (
    Cdo,
    cdo_available,
    CdoNotFoundError,
    pygrib,
    pygrib_available,
    PygribNotFoundError,
)


def create_dt_fpath(dt, root, fname, subdirs=[]):
    """
    Create filepaths from root + fname and a list of subdirectories.
    fname and subdirs will be put through dt.strftime.

    Parameters
    ----------
    dt: datetime.datetime
        date as basis for the URL
    root: string
        root of the filenpath
    fname: string
        filename to use
    subdirs: list, optional
        list of strings.
        Each element represents a subdirectory.
        For example the list ['%Y', '%m'] would lead to a URL of
        ``root/YYYY/MM/fname`` or for a dt of datetime(2000,12,31)
        ``root/2000/12/fname``

    Returns
    -------
    fpath: string
        Full filename including path
    """
    dt_subdirs = []
    for subdir in subdirs:
        dt_subdirs.append(dt.strftime(subdir))
    dt_fname = dt.strftime(fname)
    flist = [root] + dt_subdirs + [dt_fname]
    fpath = os.path.join(*flist)
    return fpath


def unzip_nc(
    input_zip,
    output_nc,
):
    """
    Unzip and merge all netcdf files downloaded from CDS. If the zip file
    contains only 1 netcdf file, it only be extracted.

    Parameters
    ----------
    input_zip: str
        Path to the downloaded zip file containing one or more (datastream)
        netcdf files.
    output_nc: str
        Path to the netcdf file to write
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        with zipfile.ZipFile(input_zip, "r") as zip_ref:
            zip_ref.extractall(tmpdir)
        ncfiles = [
            os.path.join(tmpdir, f)
            for f in os.listdir(tmpdir)
            if f.endswith(".nc")
        ]
        if len(ncfiles) == 1:
            shutil.move(ncfiles[0], output_nc)
        else:
            # Sometimes CDS returns multiple netcdf files, merge them
            ds = [xr.open_dataset(os.path.join(tmpdir, f)) for f in ncfiles]
            expvers = []
            for d in ds:
                if 'expver' in d.coords:
                    expvers.append(d.coords['expver'].values.astype(int))
            if len(expvers) > 0:
                expvers = np.array(expvers).max(axis=0)
                for d in ds:
                    d.coords['expver'] = np.array(
                        [f"{e: 04}" for e in expvers])

            ds = xr.combine_by_coords(
                ds, combine_attrs="override", compat='override')
            ds.to_netcdf(
                output_nc,
                encoding={
                    v: {
                        'zlib': True,
                        'complevel': 6
                    } for v in ds.data_vars
                })

    os.remove(input_zip)


def save_ncs_from_nc(
    input_nc,
    output_path,
    product_name,
    grid=None,
    keep_original=True,
    remap_method="bil",
    keep_prelim=True,
):
    """
    Split the downloaded netcdf file into daily files and add to folder
    structure necessary for reshuffling.

    Parameters
    ----------
    input_nc : str
        Filepath of the downloaded .nc file
    output_path : str
        Where to save the resulting netcdf files
    product_name : str
        Name of the ECMWF model (only for filename generation)
    keep_original: bool
        keep the original downloaded data too, before it is sliced into
        individual images.
    keep_prelim: bool, optional (default: True)
        True to keep preliminary data from ERA5T with a different file name, or
        False drop these files and only keep the final records.
    """
    _filename_templ = IMG_FNAME_TEMPLATE.format(
        product="{product}",
        type='AN',
        datetime=IMG_FNAME_DATETIME_FORMAT,
        ext='nc')

    with xr.open_dataset(input_nc, mask_and_scale=True) as nc_in:
        if 'valid_time' in nc_in.dims:
            nc_in = nc_in.rename_dims({"valid_time": 'time'})
        if 'valid_time' in nc_in.variables:
            nc_in = nc_in.rename_vars({"valid_time": 'time'})

        if grid is not None:
            if not cdo_available:
                raise CdoNotFoundError()
            cdo = Cdo()

            gridpath = os.path.join(output_path, "grid.txt")
            weightspath = os.path.join(output_path, "remap_weights.nc")
            if not os.path.exists(gridpath):
                with open(gridpath, "w") as f:
                    for k, v in grid.items():
                        f.write(f"{k} = {v}\n")

        for i, time in enumerate(nc_in["time"].values):
            subset = nc_in.sel({"time": time})

            # Expver identifies preliminary data
            if 'expver' in subset:
                ex = np.atleast_1d(subset['expver'].values)
                if len(ex) == 1:
                    expver = str(ex[0])
                else:
                    expver = str(ex[i])
                subset = subset.drop_vars('expver')
                try:
                    ext = EXPVER[expver]
                except KeyError:
                    ext = ''
            else:
                ext = ''

            if len(ext) > 0 and not keep_prelim:
                logging.info(f"Dropping preliminary data {time}")
                continue

            if len(ext) > 0:
                filename_templ = _filename_templ.format(product=product_name +
                                                        '-' + ext)
            else:
                filename_templ = _filename_templ.format(product=product_name)

            if 'number' in subset.variables:
                subset = subset.drop_vars('number')

            timestamp = pd.Timestamp(time).to_pydatetime()
            filepath = create_dt_fpath(
                timestamp,
                root=output_path,
                fname=filename_templ,
                subdirs=SUBDIRS,
            )

            if not os.path.exists(os.path.dirname(filepath)):
                os.makedirs(os.path.dirname(filepath))

            if grid is not None:
                if not os.path.exists(weightspath):
                    # create weights file
                    getattr(cdo, "gen" + remap_method)(
                        gridpath, input=subset, output=weightspath)
                subset = cdo.remap(
                    ",".join([gridpath, weightspath]),
                    input=subset,
                    returnXDataset=True,
                )

            # same compression for all variables
            var_encode = {"zlib": True, "complevel": 6}
            # explicitly use netcdf4 engine
            subset.to_netcdf(
                filepath,
                encoding={var: var_encode for var in subset.variables},
                engine="netcdf4")

    if not keep_original:
        os.remove(input_nc)
    if grid is not None:
        cdo.cleanTempDir()


def save_gribs_from_grib(
    input_grib,
    output_path,
    product_name,
    keep_original=True,
    keep_prelim=True,
):
    """
    Split the downloaded grib file into daily files and add to folder structure
    necessary for reshuffling.

    Parameters
    ----------
    input_grib : str
        Filepath of the downloaded .grb file
    output_path : str
        Where to save the resulting grib files
    product_name : str
        Name of the ECMWF model (only for filename generation)
    keep_original: bool
        keep the original downloaded data too, before it is sliced into
        individual images.
    keep_prelim: bool, optional (default: True)
        True to keep preliminary data from ERA5T with a different file name, or
        False drop these files and only keep the final records.
    """
    if not pygrib_available:
        raise PygribNotFoundError()
    grib_in = pygrib.open(input_grib)

    _filename_templ = IMG_FNAME_TEMPLATE.format(
        product="{product}",
        type='AN',
        datetime=IMG_FNAME_DATETIME_FORMAT,
        ext='grb')

    grib_in.seek(0)
    prev_date = None

    for grb in grib_in:
        filedate = datetime(grb["year"], grb["month"], grb["day"], grb["hour"])

        expver = grb['expver']

        try:
            ext = EXPVER[expver]
        except KeyError:
            ext = ''

        if len(ext) > 0 and not keep_prelim:
            logging.info(f"Dropping preliminary data {filedate}")
            continue

        if len(ext) > 0:
            filename_templ = _filename_templ.format(product=product_name +
                                                    '-' + ext)
        else:
            filename_templ = _filename_templ.format(product=product_name)

        filepath = create_dt_fpath(
            filedate, root=output_path, fname=filename_templ, subdirs=SUBDIRS)

        if not os.path.exists(os.path.dirname(filepath)):
            os.makedirs(os.path.dirname(filepath))

        if prev_date != filedate:  # to overwrite old files
            mode = 'wb'
            prev_date = filedate
        else:
            mode = "ab"

        with open(filepath, mode) as grb_out:
            grb_out.write(grb.tostring())

    grib_in.close()

    if not keep_original:
        os.remove(input_grib)



# -*- coding: utf-8 -*-
"""
Tests for grid generation
"""

from ecmwf_models.grid import (
    ERA_RegularImgGrid,
    get_grid_resolution,
    ERA5_RegularImgLandGrid,
)
import numpy as np


def test_global_subset():
    grid1 = ERA_RegularImgGrid(0.25)
    grid2 = ERA_RegularImgGrid(0.25, (-180, -90, 180, 90))
    np.testing.assert_array_equal(grid1.activegpis, grid2.activegpis)
    assert grid1 == grid2

    grid1 = ERA5_RegularImgLandGrid(0.25)
    grid2 = ERA5_RegularImgLandGrid(0.25, (-180, -90, 180, 90))
    np.testing.assert_array_equal(grid1.activegpis, grid2.activegpis)
    assert grid1 == grid2

    grid1 = ERA5_RegularImgLandGrid(0.25)
    grid2 = ERA5_RegularImgLandGrid(0.25, (-10, 60, 30, 90))
    assert grid1 != grid2


def test_ERA_regular_grid():
    reg_grid = ERA_RegularImgGrid(0.3)
    assert np.unique(reg_grid.activearrlat).size == 601
    assert np.unique(reg_grid.activearrlon).size == 1200
    assert get_grid_resolution(reg_grid.activearrlat,
                               reg_grid.activearrlon) == (
                                   0.3,
                                   0.3,
                               )

def test_ERA5_landgrid_025():
    grid = ERA5_RegularImgLandGrid(0.25)  # 0.25*0.25
    assert grid.get_grid_points()[0].size == 244450
    assert grid.find_nearest_gpi(16.25, 48.25) == (240545, 0.0)
    assert grid.gpi2cell(240545) == 1431


def test_ERA5_landgrid_01():
    grid = ERA5_RegularImgLandGrid(0.1)  # 0.1*0.1
    assert grid.get_grid_points()[0].size == 1544191
    assert grid.find_nearest_gpi(16.4, 48.1) == (1508564, 0.0)
    np.testing.assert_almost_equal(grid.gpi2lonlat(1508564)[0], 16.4)
    np.testing.assert_almost_equal(grid.gpi2lonlat(1508564)[1], 48.1)
    assert grid.gpi2cell(1508564) == 1431



import os
from pathlib import Path

IMG_FNAME_TEMPLATE = "{product}_{type}_{datetime}.{ext}"
IMG_FNAME_DATETIME_FORMAT = "%Y%m%d_%H%M"

# ERA5 products supported by the reader.
SUPPORTED_PRODUCTS = ['era5', 'era5-land']

CDS_API_URL = "https://cds.climate.copernicus.eu/api"

# CDSAPI_RC variable must be set or we use home dir
DOTRC = os.environ.get('CDSAPI_RC', os.path.join(Path.home(), '.cdsapirc'))

EXPVER = {'0005': 'T', '0001': ''}

SUBDIRS = ["%Y", "%j"]

try:
    import pygrib
    pygrib_available = True
except ImportError:
    pygrib_available = False
    pygrib = None

try:
    from cdo import Cdo
    cdo_available = True
except ImportError:
    cdo_available = False
    Cdo = None


class PygribNotFoundError(ModuleNotFoundError):

    def __init__(self, msg=None):
        _default_msg = ("pygrib could not be imported. "
                        "Pleas run `conda install -c conda-forge pygrib` to "
                        "install it.")
        self.msg = _default_msg if msg is None else msg


class CdoNotFoundError(ModuleNotFoundError):

    def __init__(self, msg=None):
        _default_msg = (
            "cdo and/or python-cdo not installed. "
            "Pleas run `conda install -c conda-forge cdo` and also "
            "`pip install cdo`.")
        self.msg = _default_msg if msg is None else msg



# -*- coding: utf-8 -*-
"""
Common grid definitions for ECMWF model reanalysis products (regular gridded)
"""

import numpy as np
from pygeogrids.grids import CellGrid, gridfromdims
import os
from typing import Tuple
import xarray as xr


def trafo_lon(lon):
    """
    0...360 -> 0...180...-180

    Parameters
    ----------
    lon: np.array
        Longitude array

    Returns
    -------
    lon_transformed: np.array
        Transformed longitude array
    """
    lons_gt_180 = np.where(lon > 180.)
    lon[lons_gt_180] = lon[lons_gt_180] - 360.0
    return lon


def safe_arange(start, stop, step):
    """
    Like numpy.arange, but floating point precision is kept.
    Compare: `np.arange(0, 100, 0.01)[-1]` vs `safe_arange(0, 100, 0.01)[-1]`

    Parameters
    ----------
    start: float
        Start of interval
    stop: float
        End of interval (not included)
    step: float
        Stepsize

    Returns
    -------
    arange: np.array
        Range of values in interval at the given step size / sampling
    """
    f_step = (1. / float(step))
    vals = np.arange(
        float(start) * f_step,
        float(stop) * f_step,
        float(step) * f_step)
    return vals / f_step


def get_grid_resolution(lats: np.ndarray, lons: np.ndarray) -> (float, float):
    """
    try to derive the grid resolution from given coords.
    """
    lats = np.unique(lats)
    lons = np.unique(lons)
    lats_res, lons_res = [], []

    for i, j in zip(lats[:-1], lats[1:]):
        lats_res.append(np.abs(np.abs(j) - np.abs(i)))
    lats_res = np.round(np.array(lats_res), 3)
    if not all(lats_res == lats_res[0]):
        raise ValueError("Grid not regular")
    else:
        lat_res = lats_res[0]

    for i, j in zip(lons[:-1], lons[1:]):
        lons_res.append(np.abs(np.abs(j) - np.abs(i)))
    lons_res = np.round(np.array(lons_res), 3)
    if not all(lons_res == lons_res[0]):
        raise ValueError("Grid not regular")
    else:
        lon_res = lons_res[0]

    return float(lat_res), float(lon_res)


def ERA5_RegularImgLandGrid(
    resolution: float = 0.25,
    bbox: Tuple[float, float, float, float] = None,
    cellsize: float = 5.,
) -> CellGrid:
    """
    Uses the 0.25 DEG ERA5 land mask to create a land grid of the same size,
    which also excluded Antarctica.

    Parameters
    ----------
    resolution: float, optional (default: 0.25)
        Grid resolution in degrees. Either 0.25 (ERA5) or 0.1 (ERA5-Land)
    bbox: tuple, optional (default: None)
        WGS84 (min_lon, min_lat, max_lon, max_lat)
        Values must be between -180 to 180 and -90 to 90
        bbox to cut the global grid to
    cellsize: float, optional (default: 5.)
        Chunk size of the grid

    Returns
    -------
    landgrid: CellGrid
        ERA Land grid at the given resolution, cut to the given bounding box
    """
    if resolution not in [0.25, 0.1]:
        raise ValueError("Unsupported resolution. Choose one of 0.25 or 0.1")
    try:
        ds = xr.open_dataset(
            os.path.join(
                os.path.abspath(os.path.dirname(__file__)),
                "era5",
                "land_definition_files",
                f"landmask_{resolution}_{resolution}.nc",
            ))["land"]

        ds = ds.assign_coords({'longitude': trafo_lon(ds['longitude'].values)})

        if bbox is not None:
            ds = ds.sel(latitude=slice(bbox[3], bbox[1]))
            ds = ds.isel(
                longitude=np.where(((ds['longitude'].values >= bbox[0])
                                    & (ds['longitude'].values <= bbox[2])))[0])

        land_mask = np.array(ds.values == 1.0)

    except FileNotFoundError:
        raise FileNotFoundError(
            "Land definition for this grid resolution not yet available. "
            "Please create and add it.")

    full_grid = ERA_RegularImgGrid(resolution, bbox=bbox, cellsize=cellsize)

    land_gpis = full_grid.get_grid_points()[0][land_mask.flatten()]
    land_grid = full_grid.subgrid_from_gpis(land_gpis)

    return land_grid


def ERA_RegularImgGrid(
    resolution: float = 0.25,
    bbox: Tuple[float, float, float, float] = None,
    cellsize: float = 5.,
) -> CellGrid:
    """
    Create regular cell grid for bounding box with the selected
    resolution.
    GPI 0 is at Lon: 0, Lat: 90

    Parameters
    ----------
    resolution: float, optional (default: 0.25)
        Grid resolution (in degrees) in both directions.
        Either 0.25 (ERA5) or 0.1 (ERA5-Land)
    bbox: tuple, optional (default: None)
        (min_lon, min_lat, max_lon, max_lat)
        wgs84 (Lon -180 to 180)
        bbox to cut the global grid to.
    cellsize: float, optional (default: 5.)
        Cell chunking of the grid

    Returns
    ----------
    CellGrid : CellGrid
        Regular, CellGrid with 5DEG*5DEG cells for the passed bounding box.
    """
    # to get precise coordinates...
    lon = safe_arange(0.0, 360, resolution)
    lat = safe_arange(-90, 90 + resolution, resolution)[::-1]

    # ERA grid LLC point has Lon=0
    lon = trafo_lon(lon)
    grid = gridfromdims(lon, lat, origin='top')

    grid = grid.to_cell_grid(cellsize=cellsize)

    if bbox is not None:
        # could be simpler if pygeogrids kept the gpi order...
        subgpis = grid.get_bbox_grid_points(
            latmin=bbox[1], latmax=bbox[3], lonmin=bbox[0], lonmax=bbox[2])
        sel = np.where(np.isin(grid.activegpis, subgpis))
        subgpis = grid.activegpis[sel]
        sublats, sublons = grid.activearrlat[sel], grid.activearrlon[sel]
        shape = (np.unique(sublats).size, np.unique(sublons).size)
        grid = CellGrid(sublons, sublats, grid.gpi2cell(subgpis), subgpis,
                        shape=shape)

    return grid



# -*- coding: utf-8 -*-
# The MIT License (MIT)
#
# Copyright (c) 2018, TU Wien
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
'''
Tests for transferring downloaded data to netcdf or grib files
'''
import os
import shutil
from datetime import datetime
import numpy as np
import xarray as xr
import pytest
import tempfile
import zipfile

from c3s_sm.misc import read_summary_yml

from ecmwf_models.era5.download import download_and_move
from ecmwf_models.extract import save_ncs_from_nc
from ecmwf_models.globals import (
    cdo_available,
    CdoNotFoundError
)

grid = {
    "gridtype": "lonlat",
    "xsize": 720,
    "ysize": 360,
    "xfirst": -179.75,
    "yfirst": 89.75,
    "xinc": 0.5,
    "yinc": -0.5
}


@pytest.mark.skipif(cdo_available, reason="CDO is installed")
def test_download_with_cdo_not_installed():
    with pytest.raises(CdoNotFoundError):
        with tempfile.TemporaryDirectory() as out_path:
            infile = os.path.join(
                os.path.dirname(os.path.abspath(__file__)), '..',
                "ecmwf_models-test-data", "download",
                "era5_example_downloaded_raw.nc")
            save_ncs_from_nc(
                infile, out_path, 'ERA5', grid=grid,
                keep_original=True)


def test_dry_download_nc_era5():
    with tempfile.TemporaryDirectory() as dl_path:
        dl_path = os.path.join(dl_path, 'era5')
        os.mkdir(dl_path)
        os.mkdir(os.path.join(dl_path, 'temp_downloaded'))

        thefile = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), '..',
            "ecmwf_models-test-data", "download",
            "era5_example_downloaded_raw.nc")

        assert os.path.exists(thefile)

        trgt = os.path.join(dl_path, "temp_downloaded",
                            "20100101_20100101.zip")
        with zipfile.ZipFile(trgt, 'w') as zip:
            # Add the file to the ZIP archive
            zip.write(thefile, arcname="20100101_20100101.nc")

        assert os.path.isfile(trgt)

        startdate = enddate = datetime(2010, 1, 1)
        variables, h_steps = ['swvl1', 'swvl2', 'lsm'], [0, 12]

        with pytest.warns(UserWarning, match='Dry run*'):
            download_and_move(
                dl_path,
                startdate,
                enddate,
                variables=variables,
                keep_original=False,
                h_steps=h_steps,
                grb=False,
                dry_run=True)

        cont = os.listdir(dl_path)
        assert len(cont) == 2
        assert 'overview.yml' in cont
        assert '2010' in cont

        assert os.listdir(os.path.join(dl_path, '2010')) == ['001']
        assert len(os.listdir(os.path.join(dl_path, '2010', '001'))) == 2

        should_dlfiles = [
            'ERA5_AN_20100101_0000.nc', 'ERA5_AN_20100101_1200.nc'
        ]

        assert (sorted(os.listdir(os.path.join(
            dl_path, '2010', '001'))) == sorted(should_dlfiles))

        props = read_summary_yml(dl_path)
        assert props['product'].lower() == 'era5'
        assert props['download_settings']['grb'] is False
        assert props['period_to'] == props['period_from'] == '2010-01-01'
        assert len(props['download_settings']['variables']) == len(variables)
        assert props['download_settings']['h_steps'] == h_steps


def test_dry_download_grb_era5():
    with tempfile.TemporaryDirectory() as dl_path:
        dl_path = os.path.join(dl_path, 'era5')
        os.mkdir(dl_path)
        os.mkdir(os.path.join(dl_path, 'temp_downloaded'))

        thefile = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), '..',
            "ecmwf_models-test-data", "download",
            "era5_example_downloaded_raw.grb")
        shutil.copyfile(
            thefile,
            os.path.join(dl_path, 'temp_downloaded', '20100101_20100101.grb'))

        startdate = enddate = datetime(2010, 1, 1)
        variables, h_steps = ['swvl1', 'swvl2', 'lsm'], [0, 12]

        with pytest.warns(UserWarning, match="Dry run*"):
            download_and_move(
                dl_path,
                startdate,
                enddate,
                variables=variables,
                keep_original=False,
                h_steps=h_steps,
                grb=True,
                dry_run=True)

        cont = os.listdir(dl_path)
        assert len(cont) == 2
        assert 'overview.yml' in cont
        assert '2010' in cont

        assert (os.listdir(os.path.join(dl_path, '2010')) == ['001'])
        assert (len(os.listdir(os.path.join(dl_path, '2010', '001'))) == 2)

        should_dlfiles = [
            'ERA5_AN_20100101_0000.grb', 'ERA5_AN_20100101_1200.grb'
        ]

        assert (sorted(os.listdir(os.path.join(
            dl_path, '2010', '001'))) == sorted(should_dlfiles))

        props = read_summary_yml(dl_path)
        assert props['product'].lower() == 'era5'
        assert props['download_settings']['grb'] is True
        assert props['period_to'] == props['period_from'] == '2010-01-01'
        assert len(props['download_settings']['variables']) == len(variables)
        assert props['download_settings']['h_steps'] == h_steps


@pytest.mark.skipif(not cdo_available, reason="CDO is not installed")
def test_download_nc_era5_regridding():
    with tempfile.TemporaryDirectory() as dl_path:
        dl_path = os.path.join(dl_path, 'era5')
        os.mkdir(dl_path)
        os.mkdir(os.path.join(dl_path, 'temp_downloaded'))

        thefile = os.path.join(
            os.path.dirname(os.path.abspath(__file__)), '..',
            "ecmwf_models-test-data", "download",
            "era5_example_downloaded_raw.nc")

        assert os.path.exists(thefile)

        trgt = os.path.join(dl_path, "temp_downloaded",
                            "20100101_20100101.zip")
        with zipfile.ZipFile(trgt, 'w') as zip:
            # Add the file to the ZIP archive
            zip.write(thefile, arcname="20100101_20100101.nc")

        assert os.path.isfile(trgt)

        startdate = enddate = datetime(2010, 1, 1)

        with pytest.warns(UserWarning, match="Dry run*"):
            download_and_move(
                dl_path,
                startdate,
                enddate,
                variables=['swvl1', 'swvl2', 'lsm'],
                keep_original=False,
                h_steps=[0, 12],
                grb=False,
                dry_run=True,
                grid=grid)

        cont = os.listdir(dl_path)
        assert len(cont) == 2
        assert 'overview.yml' in cont
        assert '2010' in cont

        assert (os.listdir(os.path.join(dl_path, '2010')) == ['001'])
        assert (len(os.listdir(os.path.join(dl_path, '2010', '001'))) == 2)

        should_dlfiles = [
            'ERA5_AN_20100101_0000.nc', 'ERA5_AN_20100101_1200.nc'
        ]

        assert (sorted(os.listdir(os.path.join(
            dl_path, '2010', '001'))) == sorted(should_dlfiles))

        for f in os.listdir(os.path.join(dl_path, "2010", "001")):
            ds = xr.open_dataset(os.path.join(dl_path, "2010", "001", f))
            assert ds['lon'].shape == (720,)
            assert ds['lat'].shape == (360,)
            assert np.all(np.arange(89.75, -90, -0.5) == ds.lat)
            assert np.all(np.arange(-179.75, 180, 0.5) == ds.lon)


if __name__ == '__main__':
    test_dry_download_nc_era5()



# -*- coding: utf-8 -*-

import os
import numpy.testing as nptest
from ecmwf_models.era5.img import (
    ERA5NcDs, ERA5NcImg, ERA5GrbImg, ERA5GrbDs)
import numpy as np
from datetime import datetime
from ecmwf_models.grid import ERA5_RegularImgLandGrid


def test_ERA5_nc_image_landpoints():
    fname = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), '..',
        "ecmwf_models-test-data", "ERA5", "netcdf", "2010", "001",
        'ERA5_AN_20100101_0000.nc')
    subgrid = ERA5_RegularImgLandGrid(0.25)
    dset = ERA5NcImg(
        fname,
        parameter=['swvl1', 'swvl2'],
        mask_seapoints=True,
        array_1D=True,
        subgrid=subgrid)
    data = dset.read()
    assert data.data['swvl1'].shape == (244450,)
    assert data.data['swvl2'].shape == (244450,)
    assert data.lon.shape == (244450,)
    assert data.lat.shape == (244450,)
    metadata_should = {
        'long_name': u'Volumetric soil water layer 1',
        'units': u'm**3 m**-3'
    }
    assert data.metadata['swvl1']['long_name'] == metadata_should['long_name']
    assert data.metadata['swvl1']['units'] == metadata_should['units']

    metadata_should = {
        'long_name': u'Volumetric soil water layer 2',
        'units': u'm**3 m**-3'
    }
    assert data.metadata['swvl2']['long_name'] == metadata_should['long_name']
    assert data.metadata['swvl2']['units'] == metadata_should['units']

    # data over land
    nptest.assert_allclose(data.lon[1000], -29.0, rtol=1e-5)
    nptest.assert_allclose(data.lat[1000], 82.0, rtol=1e-5)
    nptest.assert_allclose(data.data['swvl1'][1000], 0.41698003, rtol=1e-5)
    #boundaries
    nptest.assert_allclose(data.lat[0], 83.5)
    nptest.assert_allclose(data.lat[-1], -55.25)
    nptest.assert_allclose(data.lon[0], -36)


def test_ERA5_nc_image():
    fname = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), '..',
        "ecmwf_models-test-data", "ERA5", "netcdf", "2010", "001",
        'ERA5_AN_20100101_0000.nc')

    dset = ERA5NcImg(fname, parameter=['swvl1', 'swvl2'], mask_seapoints=True)
    data = dset.read()
    assert data.data['swvl1'].shape == (721, 1440)
    assert data.data['swvl2'].shape == (721, 1440)
    assert data.lon.shape == (721, 1440)
    assert data.lat.shape == (721, 1440)
    metadata_should = {
        'long_name': u'Volumetric soil water layer 1',
        'units': u'm**3 m**-3'
    }
    assert data.metadata['swvl1']['long_name'] == metadata_should['long_name']
    assert data.metadata['swvl1']['units'] == metadata_should['units']

    metadata_should = {
        'long_name': u'Volumetric soil water layer 2',
        'units': u'm**3 m**-3'
    }
    assert data.metadata['swvl2']['long_name'] == metadata_should['long_name']
    assert data.metadata['swvl2']['units'] == metadata_should['units']

    # data over land
    nptest.assert_allclose(data.lon[168, 60], 15.0, rtol=1e-5)
    nptest.assert_allclose(data.lat[168, 60], 48.0, rtol=1e-5)
    nptest.assert_allclose(data.data['swvl1'][168, 60], 0.402825, rtol=1e-5)
    nptest.assert_allclose(data.data['swvl2'][168, 60], 0.390512, rtol=1e-5)
    # data over water
    nptest.assert_allclose(data.lon[400, 320], 80.0, rtol=1e-5)
    nptest.assert_allclose(data.lat[400, 320], -10.0, rtol=1e-5)
    nptest.assert_allclose(data.data['swvl1'][400, 320], np.nan)
    nptest.assert_allclose(data.data['swvl2'][400, 320], np.nan)

    #boundaries
    nptest.assert_allclose(data.lat[0, 0], 90.0)
    nptest.assert_allclose(data.lat[-1, 0], -90.0)
    nptest.assert_allclose(data.lon[0, 0], 0.0)
    nptest.assert_allclose(data.lon[0, 720], 180.0)  # middle of image


def test_ERA5_nc_image_1d():
    fname = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), '..',
        "ecmwf_models-test-data", "ERA5", "netcdf", "2010", "001",
        'ERA5_AN_20100101_0000.nc')

    dset = ERA5NcImg(
        fname,
        parameter=['swvl1', 'swvl2'],
        mask_seapoints=True,
        array_1D=True)
    data = dset.read()
    assert data.data['swvl1'].shape == (721 * 1440,)
    assert data.data['swvl2'].shape == (721 * 1440,)
    assert data.lon.shape == (721 * 1440,)
    assert data.lat.shape == (721 * 1440,)
    metadata_should = {
        'long_name': u'Volumetric soil water layer 1',
        'units': u'm**3 m**-3'
    }
    assert data.metadata['swvl1']['long_name'] == metadata_should['long_name']
    assert data.metadata['swvl1']['units'] == metadata_should['units']

    metadata_should = {
        'long_name': u'Volumetric soil water layer 2',
        'units': u'm**3 m**-3'
    }
    assert data.metadata['swvl2']['long_name'] == metadata_should['long_name']
    assert data.metadata['swvl2']['units'] == metadata_should['units']

    # data over land
    nptest.assert_allclose(data.lon[168 * 1440 + 60], 15.0, rtol=1e-5)
    nptest.assert_allclose(data.lat[168 * 1440 + 60], 48.0, rtol=1e-5)
    nptest.assert_allclose(
        data.data['swvl1'][168 * 1440 + 60], 0.402825, rtol=1e-5)
    nptest.assert_allclose(
        data.data['swvl2'][168 * 1440 + 60], 0.390512, rtol=1e-5)
    # data over water
    nptest.assert_allclose(data.lon[400 * 1440 + 320], 80.0, rtol=1e-4)
    nptest.assert_allclose(data.lat[400 * 1440 + 320], -10.0, rtol=1e-4)
    nptest.assert_allclose(data.data['swvl1'][400 * 1440 + 320], np.nan)
    nptest.assert_allclose(data.data['swvl2'][400 * 1440 + 320], np.nan)

    #boundaries
    nptest.assert_allclose(data.lat[0], 90.0)
    nptest.assert_allclose(data.lat[-1], -90.0)
    nptest.assert_allclose(data.lon[0], 0.0)
    nptest.assert_allclose(data.lon[len(data.lon) // 2],
                           180.0)  # middle of image


def test_ERA5_grb_image():
    fname = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), '..',
        "ecmwf_models-test-data", "ERA5", "grib", "2010", "001",
        'ERA5_AN_20100101_0000.grb')

    dset = ERA5GrbImg(
        fname,
        parameter=['swvl1', 'swvl2'],
        mask_seapoints=True,
        array_1D=False)
    data = dset.read()
    assert data.data['swvl1'].shape == (721, 1440)
    assert data.data['swvl2'].shape == (721, 1440)
    assert data.lon.shape == (721, 1440)
    assert data.lat.shape == (721, 1440)
    metadata_should = {
        'long_name': u'Volumetric soil water layer 1',
        'units': u'm**3 m**-3'
    }
    assert data.metadata['swvl1']['long_name'] == metadata_should['long_name']
    assert data.metadata['swvl1']['units'] == metadata_should['units']

    metadata_should = {
        'long_name': u'Volumetric soil water layer 2',
        'units': u'm**3 m**-3'
    }
    assert data.metadata['swvl2']['long_name'] == metadata_should['long_name']
    assert data.metadata['swvl2']['units'] == metadata_should['units']

    # data over land
    nptest.assert_allclose(data.lon[168, 60], 15.0, rtol=1e-5)
    nptest.assert_allclose(data.lat[168, 60], 48.0, rtol=1e-5)
    nptest.assert_allclose(data.data['swvl1'][168, 60], 0.402824, rtol=1e-4)
    nptest.assert_allclose(data.data['swvl2'][168, 60], 0.390514, rtol=1e-4)
    # data over water
    nptest.assert_allclose(data.lon[400, 320], 80.0, rtol=1e-5)
    nptest.assert_allclose(data.lat[400, 320], -10.0, rtol=1e-5)
    nptest.assert_allclose(data.data['swvl1'][400, 320], np.nan)
    nptest.assert_allclose(data.data['swvl2'][400, 320], np.nan)

    # boundaries
    nptest.assert_allclose(data.lat[0, 0], 90.0)
    nptest.assert_allclose(data.lat[-1, 0], -90.0)
    nptest.assert_allclose(data.lon[0, 0], 0.0)
    nptest.assert_allclose(data.lon[0, 720], 180.0)  # middle of image


def test_ERA5_grb_image_1d():
    fname = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), '..',
        "ecmwf_models-test-data", "ERA5", "grib", "2010", "001",
        'ERA5_AN_20100101_0000.grb')

    dset = ERA5GrbImg(
        fname,
        parameter=['swvl1', 'swvl2'],
        mask_seapoints=True,
        array_1D=True)
    data = dset.read()
    assert data.data['swvl1'].shape == (721 * 1440,)
    assert data.data['swvl2'].shape == (721 * 1440,)
    assert data.lon.shape == (721 * 1440,)
    assert data.lat.shape == (721 * 1440,)
    metadata_should = {
        'long_name': u'Volumetric soil water layer 1',
        'units': u'm**3 m**-3'
    }
    assert data.metadata['swvl1']['long_name'] == metadata_should['long_name']
    assert data.metadata['swvl1']['units'] == metadata_should['units']

    metadata_should = {
        'long_name': u'Volumetric soil water layer 2',
        'units': u'm**3 m**-3'
    }
    assert data.metadata['swvl2']['long_name'] == metadata_should['long_name']
    assert data.metadata['swvl2']['units'] == metadata_should['units']

    # data over land
    nptest.assert_allclose(data.lon[168 * 1440 + 60], 15.0, rtol=1e-5)
    nptest.assert_allclose(data.lat[168 * 1440 + 60], 48.0, rtol=1e-5)
    nptest.assert_allclose(
        data.data['swvl1'][168 * 1440 + 60], 0.402824, rtol=1e-4)
    nptest.assert_allclose(
        data.data['swvl2'][168 * 1440 + 60], 0.390514, rtol=1e-4)
    # data over water
    nptest.assert_allclose(data.lon[400 * 1440 + 320], 80.0, rtol=1e-5)
    nptest.assert_allclose(data.lat[400 * 1440 + 320], -10.0, rtol=1e-5)
    nptest.assert_allclose(data.data['swvl1'][400 * 1440 + 320], np.nan)
    nptest.assert_allclose(data.data['swvl2'][400 * 1440 + 320], np.nan)

    # boundaries
    nptest.assert_allclose(data.lat[0], 90.0)
    nptest.assert_allclose(data.lat[-1], -90.0)
    nptest.assert_allclose(data.lon[0], 0.0)
    nptest.assert_allclose(data.lon[len(data.lon) // 2],
                           180.0)  # middle of image


def test_ERA5_nc_ds():
    root_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), '..',
        "ecmwf_models-test-data", "ERA5", "netcdf")

    tstamps_should = [datetime(2010, 1, 1), datetime(2010, 1, 1, 12)]

    ds = ERA5NcDs(
        root_path,
        parameter=['swvl1', 'swvl2'],
        array_1D=True,
        mask_seapoints=True,
        h_steps=[0, 12])

    for data, tstamp_should in zip(
            ds.iter_images(datetime(2010, 1, 1), datetime(2010, 1, 1)),
            tstamps_should):
        assert data.data['swvl1'].shape == (721 * 1440,)
        assert data.data['swvl2'].shape == (721 * 1440,)
        assert data.lon.shape == (721 * 1440,)
        assert data.lat.shape == (721 * 1440,)
        assert data.timestamp == tstamp_should
        metadata_should = {
            'swvl1': {
                'long_name': u'Volumetric soil water layer 1',
                'units': u'm**3 m**-3'
            },
            'swvl2': {
                'long_name': u'Volumetric soil water layer 2',
                'units': u'm**3 m**-3'
            }
        }
        for var in ['swvl1', 'swvl2']:
            assert data.metadata[var]['long_name'] == metadata_should[var][
                'long_name']
            assert data.metadata[var]['units'] == metadata_should[var]['units']

        nptest.assert_allclose(data.lat[0], 90.0)
        nptest.assert_allclose(data.lat[-1], -90.0)
        nptest.assert_allclose(data.lon[0], 0.0)
        nptest.assert_allclose(data.lon[720], 180.0)  # middle of image


def test_ERA5_grb_ds():
    root_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), '..',
        "ecmwf_models-test-data", "ERA5", "grib")

    tstamps_should = [datetime(2010, 1, 1), datetime(2010, 1, 1, 12)]

    ds = ERA5GrbDs(
        root_path,
        parameter=['swvl1', 'swvl2'],
        array_1D=True,
        mask_seapoints=True,
        h_steps=[0, 12])

    for data, tstamp_should in zip(
            ds.iter_images(datetime(2010, 1, 1), datetime(2010, 1, 1)),
            tstamps_should):
        assert data.data['swvl1'].shape == (721 * 1440,)
        assert data.data['swvl2'].shape == (721 * 1440,)
        assert data.lon.shape == (721 * 1440,)
        assert data.lat.shape == (721 * 1440,)
        assert data.timestamp == tstamp_should
        metadata_should = {
            'swvl1': {
                'long_name': u'Volumetric soil water layer 1',
                'units': u'm**3 m**-3'
            },
            'swvl2': {
                'long_name': u'Volumetric soil water layer 2',
                'units': u'm**3 m**-3'
            }
        }
        for var in ['swvl1', 'swvl2']:
            assert data.metadata[var]['long_name'] == metadata_should[var][
                'long_name']
            assert data.metadata[var]['units'] == metadata_should[var]['units']

        nptest.assert_allclose(data.lat[0], 90.0)
        nptest.assert_allclose(data.lat[-1], -90.0)
        nptest.assert_allclose(data.lon[0], 0.0)
        nptest.assert_allclose(data.lon[720], 180.0)  # middle of image

if __name__ == '__main__':
    test_ERA5_grb_image()


# -*- coding: utf-8 -*-
"""
Image to time series conversion tools.
"""
import warnings

import os
import pandas as pd
import numpy as np
from datetime import time, datetime, timedelta
from typing import Tuple, Union
from pygeogrids.grids import CellGrid
import inspect

from repurpose.img2ts import Img2Ts
from c3s_sm.misc import update_ts_summary_file, read_summary_yml

from ecmwf_models.era5.img import ERA5NcDs, ERA5GrbDs
from ecmwf_models.grid import ERA_RegularImgGrid, ERA5_RegularImgLandGrid
from ecmwf_models.utils import (parse_filetype, parse_product,
                                get_first_last_image_date)


class Reshuffler:
    def __init__(self, input_root, outputpath, variables=None,
                 h_steps=(0, 6, 12, 18), product=None, land_points=False):
        """

        Parameters
        ----------
        input_root: str
            Input path where ERA image data was downloaded to.
        outputpath : str
            Output path, where the reshuffled netcdf time series are stored.
        variables: tuple or str
            Variables to read from the passed images and convert into time
            series format.
        h_steps : list or tuple, optional (default: (0, 6, 12, 18))
            Hours at which images are read for each day and used for
            reshuffling, therefore this defines the sub-daily temporal
            resolution of the time series that are generated.
        product : str, optional (default: None)
            Either era5 or era5-land, if None is passed we guess the product
            from the downloaded image files.
        land_points: bool, optional (default: False)
            Reshuffle only land points. Uses the ERA5 land mask to create a
            land grid.
            The land grid is fixed to 0.25*0.25 or 0.1*0.1 deg for now.
        """

        self.input_root = input_root
        self.outputpath = outputpath
        self.land_points = land_points
        self.variables = variables
        self.h_steps = h_steps

        self.filetype = parse_filetype(input_root)
        self.product = product or parse_product(input_root)

    def find_first_last_file_date(self) -> Tuple[str, str]:
        """
        Derive time stamp of the first and last available image
        """
        first_date = get_first_last_image_date(self.input_root,
                                               start_from_last=False)
        last_date = get_first_last_image_date(self.input_root,
                                              start_from_last=True)

        return first_date, last_date

    def load_grid(self, bbox: Tuple = None, cellsize=5.) -> CellGrid:
        """
        Generate ERA5 and ERA5-Land grid in the given bounding box.
        Ensures that GPIs are consistent with global grid.
        """
        if self.land_points:
            if self.product == "era5":
                grid = ERA5_RegularImgLandGrid(resolution=0.25, bbox=bbox,
                                               cellsize=cellsize)
            elif self.product == "era5-land":
                grid = ERA5_RegularImgLandGrid(resolution=0.1, bbox=bbox,
                                               cellsize=cellsize)
            else:
                raise NotImplementedError(
                    self.product, "Land grid not implemented for product.")
        else:
            if self.product.lower() in ['era5-land', 'era5-land-t']:
                grid = ERA_RegularImgGrid(resolution=0.1, bbox=bbox,
                                          cellsize=cellsize)
            elif self.product.lower() in ['era5', 'era5-t']:
                grid = ERA_RegularImgGrid(resolution=0.25, bbox=bbox,
                                          cellsize=cellsize)
            else:
                raise NotImplementedError(self.product,
                                          "Grid not implemented for product.")

        return grid

    def get_img_reader(self,
                       grid: Union[CellGrid, None],
                       h_steps: tuple) -> Union[ERA5GrbDs, ERA5NcDs]:
        """
        Set up the Multi Image reader class
        """
        if self.filetype == "grib":
            input_dataset = ERA5GrbDs(
                root_path=self.input_root,
                parameter=self.variables,
                subgrid=grid,
                array_1D=True,
                h_steps=h_steps,
                product=self.product,
                mask_seapoints=False,
            )
        elif self.filetype == "netcdf":
            input_dataset = ERA5NcDs(
                root_path=self.input_root,
                parameter=self.variables,
                subgrid=grid,
                array_1D=True,
                h_steps=h_steps,
                product=self.product,
                mask_seapoints=False,
            )
        else:
            raise Exception("Unknown file format")

        return input_dataset

    def reshuffle(self, startdate=None, enddate=None, bbox=None,
                  cellsize=5.0, imgbuffer=50):
        """
        Reshuffle method applied to ERA images for conversion into netcdf time
        series format.
        Note: ERA5 and ERA5-Land files are preferred over their T-counterpart,
        in case that multiple files for a time stamp are present!

        Parameters
        ----------
        startdate : str
            Start date, from which images are read and time series are
            generated.
            Format YYYY-mm-dd
        enddate : str, optional (default: None)
            End date, from which images are read and time series are generated.
            Format YYYY-mm-dd. If None is passed, then the last available image
            date is used.
        variables: tuple or str
            Variables to read from the passed images and convert into time
            series format.
        product : str, optional (default: None)
            Either era5 or era5-land, if None is passed we guess the product
            from the downloaded image files.
        bbox: tuple optional (default: None)
            (min_lon, min_lat, max_lon, max_lat) - wgs84.
            To load only a subset of the global grid / file.
        cellsize: float, optional (default: 5.0)
            Cell chunking of the time series in degrees.
        imgbuffer: int, optional (default: 50)
            How many images to read at once before writing time series.
            This number affects how many images are stored in memory and should
            be chosen according to the available amount of memory and the size
            of a single image.
        """

        if (startdate is None) or (enddate is None):
            _first, _last = self.find_first_last_file_date()
            startdate = _first if startdate is None else startdate
            enddate = _last if enddate is None else enddate

        startdate = pd.to_datetime(startdate).to_pydatetime().date()
        enddate = pd.to_datetime(enddate).to_pydatetime().date()

        if startdate > enddate:
            warnings.warn(
                f"Start date {startdate} is greater than end date {enddate}. "
                f"Abort."
            )
            return

        # get time series attributes from first day of data.
        first_date_time = datetime.combine(
            startdate, time(self.h_steps[0], 0))

        if bbox is None:
            # If a subset was downloaded, otherwise repurpose would resample
            input_dataset = self.get_img_reader(None, self.h_steps)
            _img = input_dataset.read(first_date_time)
            bbox = [round(float(_img.lon.min()), 3),
                    round(float(_img.lat.min()), 3),
                    round(float(_img.lon.max()), 3),
                    round(float(_img.lat.max()), 3)]

        grid = self.load_grid(bbox, cellsize=cellsize)
        input_dataset = self.get_img_reader(grid, self.h_steps)

        global_attr = {
            f"product": f"{self.product.upper()} (from {self.filetype})"
        }

        # get time series attributes from first day of data.
        data = input_dataset.read(first_date_time)
        ts_attributes = data.metadata

        props = {
            'product': self.product,
            'filetype': self.filetype,
            'last_update': datetime.now()
        }

        kwargs = {
            'input_root': self.input_root,
            'outputpath': self.outputpath,
            'variables':
                None if self.variables is None else list(self.variables),
            'land_points': self.land_points,
            'startdate': startdate,
            'enddate': enddate,
            'h_steps': list(self.h_steps),
            'bbox': None if bbox is None else list(bbox),
            'imgbuffer': imgbuffer,
            'cellsize': cellsize,
        }

        props["img2ts_kwargs"] = kwargs

        os.makedirs(self.outputpath, exist_ok=True)

        reshuffler = Img2Ts(
            input_dataset=input_dataset,
            outputpath=self.outputpath,
            startdate=str(startdate),
            enddate=str(enddate),
            input_grid=grid,
            imgbuffer=imgbuffer,
            ts_dtypes=np.dtype("float32"),
            global_attr=global_attr,
            zlib=True,
            unlim_chunksize=1000,
            ts_attributes=ts_attributes,
            backend='multiprocessing',
            n_proc=1,
        )

        reshuffler.calc()

        update_ts_summary_file(self.outputpath, props)


def extend_ts(ts_path, **img2ts_kwargs):
    """
    Append any new data from the image path to the time series data.
    This function is only applied to time series file that were created
    using the img2ts function.
    This will use the start from the previously written metadata,
    and process only parameters that are already present in the time series
    files.

    Parameters
    ----------
    ts_path: str
        Directory containing time series files to extend. It is also
        expected that there is a `overview.yml` file in this directory.
    img2ts_kwargs:
        All kwargs are optional, if they are not given, we use them from the
        previously created `overview.yml` file. If they are passed here,
        they will override the values from the yml file.
    """
    if not os.path.isfile(os.path.join(ts_path, 'overview.yml')):
        raise FileNotFoundError(
            "No overview.yml file found in the passed directory. "
            "This file is required to use the same settings to extend an "
            "existing record. NOTE: Use the `era5 download` or "
            "`era5land download` programs first.")

    ts_props = read_summary_yml(ts_path)

    for k, v in ts_props['img2ts_kwargs'].items():
        if k not in img2ts_kwargs:
            if k == 'startdate':
                old_enddate = pd.to_datetime(
                    ts_props["img2ts_kwargs"]["enddate"]).to_pydatetime()
                startdate = old_enddate + timedelta(days=1)
                img2ts_kwargs['startdate'] = startdate
            elif k in ['enddate', 'outputpath']:  # fields from yml to ignore
                continue
            else:
                img2ts_kwargs[k] = v

    if 'enddate' not in img2ts_kwargs:
        img2ts_kwargs['enddate'] = None  # All available images

    reshuffler_kwargs = {}
    for n in inspect.getfullargspec(Reshuffler).args[1:]:
        if n in img2ts_kwargs:
            reshuffler_kwargs[n] = img2ts_kwargs[n]

    reshuffle_kwargs = {}
    for n in inspect.getfullargspec(Reshuffler.reshuffle).args[1:]:
        if n in img2ts_kwargs:
            reshuffle_kwargs[n] = img2ts_kwargs[n]

    reshuffler = Reshuffler(outputpath=ts_path, **reshuffler_kwargs)
    reshuffler.reshuffle(**reshuffle_kwargs)
